"""
.. module:: samplers.mcmc

:Synopsis: Blocked fast-slow Metropolis sampler (Lewis 1304.4473)
:Author: Antony Lewis (for the CosmoMC sampler, wrapped for cobaya by Jesus Torrado)
"""

# Global
import re
import datetime
from itertools import chain
from typing import Sequence, Optional, Callable, Union, TYPE_CHECKING
import numpy as np
from pandas import DataFrame

# Local
from cobaya.sampler import CovmatSampler
from cobaya.mpi import get_mpi_size, share_mpi, more_than_one_process, is_main_process, \
    sync_processes
from cobaya.collection import SampleCollection, OneSamplePoint, apply_temperature_cov, \
    remove_temperature_cov, remove_temperature
from cobaya.conventions import OutPar, Extension, line_width, get_version
from cobaya.typing import empty_dict
from cobaya.samplers.mcmc.proposal import BlockedProposer
from cobaya.log import LoggedError, always_stop_exceptions
from cobaya.tools import get_external_function, NumberWithUnits, load_DataFrame
from cobaya.yaml import yaml_dump_file
from cobaya.model import LogPosterior
from cobaya import mpi

# Avoid importing GetDist if not necessary
if TYPE_CHECKING:
    from getdist import MCSamples


# Suppresses warnings about first defining attrs outside __init__
# pylint: disable=attribute-defined-outside-init


class MCMC(CovmatSampler):
    r"""
    Adaptive, speed-hierarchy-aware MCMC sampler (adapted from CosmoMC)
    \cite{Lewis:2002ah,Lewis:2013hha}.
    """

    _at_resume_prefer_new = CovmatSampler._at_resume_prefer_new + [
        "burn_in", "callback_function", "callback_every", "max_tries", "output_every",
        "learn_every", "learn_proposal_Rminus1_max", "learn_proposal_Rminus1_max_early",
        "learn_proposal_Rminus1_min", "max_samples", "Rminus1_stop", "Rminus1_cl_stop",
        "Rminus1_cl_level", "covmat", "covmat_params"]
    _at_resume_prefer_old = CovmatSampler._at_resume_prefer_new + [
        "proposal_scale", "blocking"]
    _prior_rejections: int = 0
    file_base_name = 'mcmc'

    # instance variables from yaml
    burn_in: NumberWithUnits
    learn_every: NumberWithUnits
    output_every: NumberWithUnits
    callback_every: NumberWithUnits
    temperature: float
    max_tries: NumberWithUnits
    max_samples: int
    drag: bool
    callback_function: Optional[Callable]
    blocking: Optional[Sequence]
    proposal_scale: float
    learn_proposal: bool
    learn_proposal_Rminus1_max: float
    learn_proposal_Rminus1_max_early: float
    Rminus1_cl_level: float
    Rminus1_stop: float
    Rminus1_cl_stop: float
    Rminus1_single_split: int
    learn_proposal_Rminus1_min: float
    measure_speeds: bool
    oversample_thin: int
    oversample_power: float

    def set_instance_defaults(self):
        """Ensure that checkpoint attributes are initialized correctly."""
        super().set_instance_defaults()
        # checkpoint variables
        self.converged = False
        self.mpi_size = None
        self.Rminus1_last = np.inf

    def initialize(self):
        """
        Initializes the sampler: creates the proposal distribution and draws the initial
        sample.
        """
        if not self.model.prior.d():
            raise LoggedError(self.log, "No parameters being varied for sampler")
        self.log.debug("Initializing")
        if self.callback_every is None:
            self.callback_every = self.learn_every
        self._quants_d_units = []
        for q in ["max_tries", "learn_every", "callback_every", "burn_in"]:
            number = NumberWithUnits(getattr(self, q), "d", dtype=int)
            self._quants_d_units.append(number)
            setattr(self, q, number)
        self.output_every = NumberWithUnits(self.output_every, "s", dtype=int)
        if self.temperature is None:
            self.temperature = 1
        elif self.temperature < 1:
            self.mpi_warning("Sampling temperatures <1 can lead to innacurate inference.")
        if is_main_process():
            if self.output.is_resuming() and (
                    max(self.mpi_size or 0, 1) != mpi.size()):
                raise LoggedError(
                    self.log,
                    "Cannot resume a run with a different number of chains: "
                    "was %d and now is %d.", max(self.mpi_size or 0, 1), mpi.size())
        sync_processes()
        # One collection per MPI process: `name` is the MPI rank + 1
        name = str(1 + mpi.rank())
        self.collection = SampleCollection(
            self.model, self.output, name=name, resuming=self.output.is_resuming(),
            temperature=self.temperature, sample_type="mcmc")
        self.current_point = OneSamplePoint(self.model)
        # Use standard MH steps by default
        self.get_new_sample = self.get_new_sample_metropolis
        # Prepare callback function
        if self.callback_function:
            self.callback_function_callable = (
                get_external_function(self.callback_function))
        # Useful for getting last points added inside callback function
        self.last_point_callback = 0
        self.i_learn = 1
        # Monitoring/restore progress
        if is_main_process():
            cols = ["N", "timestamp", "acceptance_rate", "Rminus1", "Rminus1_cl"]
            self.progress = DataFrame(columns=cols)
            if self.output and not self.output.is_resuming():
                header_fmt = {"N": 6 * " " + "N", "timestamp": 17 * " " + "timestamp"}
                with open(self.progress_filename(), "w",
                          encoding="utf-8") as progress_file:
                    progress_file.write("# " + " ".join(
                        [header_fmt.get(col, ((7 + 8) - len(col)) * " " + col) for col in
                         self.progress.columns]) + "\n")
        # Get first point, to be discarded -- not possible to determine its weight
        # Still, we need to compute derived parameters, since, as the proposal "blocked",
        # we may be saving the initial state of some block.
        # NB: if resuming but nothing was written (burn-in not finished): re-start
        if self.output.is_resuming() and len(self.collection):
            last = len(self.collection) - 1
            initial_point = (self.collection[self.collection.sampled_params]
                             .iloc[last]).to_numpy(dtype=np.float64, copy=True)
            results = LogPosterior(
                logpost=-remove_temperature(
                    self.collection[OutPar.minuslogpost].iloc[last], self.temperature),
                logpriors=-(self.collection[self.collection.minuslogprior_names]
                            .iloc[last].to_numpy(dtype=np.float64, copy=True)),
                loglikes=-0.5 * (self.collection[self.collection.chi2_names]
                                 .iloc[last].to_numpy(dtype=np.float64, copy=True)),
                derived=(self.collection[self.collection.derived_params].iloc[last]
                         .to_numpy(dtype=np.float64, copy=True)))
        else:
            # NB: max_tries adjusted to dim instead of #cycles (blocking not computed yet)
            self.max_tries.set_scale(self.model.prior.d())
            self.log.info("Getting initial point... (this may take a few seconds)")
            initial_point, results = \
                self.model.get_valid_point(max_tries=min(self.max_tries.value, int(1e7)),
                                           random_state=self._rng)
            # If resuming but no existing chain, assume failed run and ignore blocking
            # if speeds measurement requested
            if self.output.is_resuming() and not self.collection \
               and self.measure_speeds:
                self.blocking = None
            if self.measure_speeds and self.blocking:
                self.mpi_warning(
                    "Parameter blocking manually fixed: speeds will not be measured.")
            elif self.measure_speeds:
                n = None if self.measure_speeds is True else int(self.measure_speeds)
                self.model.measure_and_set_speeds(n=n, discard=0, random_state=self._rng)
        self.current_point.add(initial_point, results)
        self.log.info("Initial point: %s", self.current_point)
        # Set up blocked proposer
        self.set_proposer_blocking()
        self.set_proposer_initial_covmat(load=True)
        # Max #(learn+convergence checks) to wait,
        # in case one process dies/hangs without raising error
        self.been_waiting = 0
        self.max_waiting = max(50, self.max_tries.unit_value)
        # Burning-in countdown -- the +1 accounts for the initial point (always accepted)
        self.burn_in_left = self.burn_in.value * self.current_point.output_thin + 1
        self._msg_ready = ("Ready to check convergence" +
                           (" and learn a new proposal covmat"
                            if self.learn_proposal else ""))

        # Initial dummy checkpoint
        # (needed when 1st "learn point" not reached in prev. run)
        self.write_checkpoint()

    @property
    def i_last_slow_block(self):
        """Block-index of the last block considered slow, if binary fast/slow split."""
        if self.drag:
            return next(i for i, o in enumerate(self.oversampling_factors) if o != 1) - 1
        self.log.warning("`i_last_slow_block` is only well defined when dragging.")
        return 0

    @property
    def slow_blocks(self):
        """Parameter blocks which are considered slow, in binary fast/slow splits."""
        return self.blocks[:1 + self.i_last_slow_block]

    @property
    def slow_params(self):
        """Parameters which are considered slow, in binary fast/slow splits."""
        return list(chain(*self.slow_blocks))

    @property
    def n_slow(self):
        """Number of parameters which are considered slow, in binary fast/slow splits."""
        return len(self.slow_params)

    @property
    def fast_blocks(self):
        """Parameter blocks which are considered fast, in binary fast/slow splits."""
        return self.blocks[self.i_last_slow_block + 1:]

    @property
    def fast_params(self):
        """Parameters which are considered fast, in binary fast/slow splits."""
        return list(chain(*self.fast_blocks))

    @property
    def n_fast(self):
        """Number of parameters which are considered fast, in binary fast/slow splits."""
        return len(self.fast_params)

    def get_acceptance_rate(self, first=0, last=None):
        """
        Computes the current acceptance rate, optionally only for ``[first:last]``
        subchain.
        """
        return (((last or self.n()) - (first or 0)) /
                self.collection[OutPar.weight][first:last].sum())

    def set_proposer_blocking(self):
        """Sets up the blocked proposer."""
        if self.blocking:
            # Includes the case in which we are resuming
            self.blocks, self.oversampling_factors = \
                self.model.check_blocking(self.blocking)
        else:
            self.blocks, self.oversampling_factors = \
                self.model.get_param_blocking_for_sampler(
                    oversample_power=self.oversample_power, split_fast_slow=self.drag)
        # Turn off dragging if one block, or if speed differences < 2x, or no differences
        if self.drag:
            if len(self.blocks) == 1:
                self.drag = False
                self.mpi_warning(
                    "Dragging disabled: not possible if there is only one block.")
            if max(self.oversampling_factors) / min(self.oversampling_factors) < 2:
                self.drag = False
                self.mpi_warning("Dragging disabled: speed ratios < 2.")
        if self.drag:
            # The definition of oversample_power=1 as spending the same amount of time in
            # the slow and fast block would suggest a 1/2 factor here, but this additional
            # factor of 2 w.r.t. oversampling should produce an equivalent exploration
            # efficiency.
            self.drag_interp_steps = int(
                np.round(self.oversampling_factors[self.i_last_slow_block + 1] *
                         self.n_fast / self.n_slow))
            if self.drag_interp_steps < 2:
                self.drag = False
                self.mpi_warning("Dragging disabled: "
                                 "speed ratio and fast-to-slow ratio not large enough.")
        # Define proposer and other blocking-related quantities
        if self.drag:
            self.get_new_sample = self.get_new_sample_dragging
            self.mpi_info("Dragging with number of interpolating steps:")
            max_width = len(str(self.drag_interp_steps))
            self.mpi_info("* %" + "%d" % max_width + "d : %r", 1, self.slow_blocks)
            self.mpi_info("* %" + "%d" % max_width + "d : %r",
                          self.drag_interp_steps, self.fast_blocks)
        elif np.any(np.array(self.oversampling_factors) > 1):
            self.mpi_info("Oversampling with factors:")
            max_width = len(str(max(self.oversampling_factors)))
            for f, b in zip(self.oversampling_factors, self.blocks):
                self.mpi_info("* %" + "%d" % max_width + "d : %r", f, b)
            if self.oversample_thin:
                self.current_point.output_thin = int(
                    np.round(
                        sum(
                            len(b) * o for b, o in zip(
                                self.blocks,
                                self.oversampling_factors,
                            )
                        ) / self.model.prior.d()
                    )
                )
        # Save blocking in updated info, in case we want to resume
        self._updated_info["blocking"] = list(zip(self.oversampling_factors, self.blocks))
        sampled_params_list = list(self.model.parameterization.sampled_params())
        blocks_indices = [[sampled_params_list.index(p) for p in b] for b in self.blocks]
        self.proposer = BlockedProposer(
            blocks_indices, self._rng,
            oversampling_factors=self.oversampling_factors,
            i_last_slow_block=(self.i_last_slow_block if self.drag else None),
            proposal_scale=self.proposal_scale)
        # Cycle length, taking into account oversampling/dragging
        if self.drag:
            self.cycle_length = self.n_slow
        else:
            self.cycle_length = sum(len(b) * o for b, o in
                                    zip(blocks_indices, self.oversampling_factors))
        self.mpi_debug(
            "Cycle length in steps: %r", self.cycle_length)
        for number in self._quants_d_units:
            number.set_scale(self.cycle_length // self.current_point.output_thin)

    def set_proposer_initial_covmat(self, load=False):
        """Creates/loads an initial covariance matrix and sets it in the Proposer."""
        if load:
            # Build the initial covariance matrix of the proposal, or load from checkpoint
            self._initial_covmat, where_nan = self._load_covmat(
                prefer_load_old=self.output.is_resuming())
            if np.any(where_nan) and self.learn_proposal:
                # We want to start learning the covmat earlier.
                self.mpi_info("Covariance matrix " +
                              ("not present" if np.all(where_nan) else "not complete") +
                              ". We will start learning the covariance of the proposal "
                              "earlier: R-1 = %g (would be %g if all params loaded).",
                              self.learn_proposal_Rminus1_max_early,
                              self.learn_proposal_Rminus1_max)
                self.learn_proposal_Rminus1_max = self.learn_proposal_Rminus1_max_early
            self.mpi_debug(
                "Sampling with covmat:\n%s",
                DataFrame(self._initial_covmat,
                          columns=self.model.parameterization.sampled_params(),
                          index=self.model.parameterization.sampled_params()).to_string(
                    line_width=line_width))
        self.proposer.set_covariance(
            apply_temperature_cov(self._initial_covmat, self.temperature))

    def _get_last_nondragging_block(self, blocks, speeds):
        # blocks and speeds are already sorted
        log_differences = np.zeros(len(blocks) - 1)
        for i in range(len(blocks) - 1):
            log_differences[i] = (np.log(np.min(speeds[:i + 1])) -
                                  np.log(np.min(speeds[i + 1:])))
        i_max = np.argmin(log_differences)
        return i_max

    def run(self):
        """
        Runs the sampler.
        """
        self.mpi_info("Sampling!" +
                      (" (NB: no accepted step will be saved until %d burn-in samples " %
                       self.burn_in.value + "have been obtained)"
                       if self.burn_in.value else ""))
        self.n_steps_raw = 0
        last_output: float = 0
        last_n = self.n()
        state_check_every = 1
        with mpi.ProcessState(self) as state:
            while last_n < self.max_samples and not self.converged:
                self.get_new_sample()
                self.n_steps_raw += 1
                if self.output_every.unit:
                    # if output_every in sec, print some info
                    # and dump at fixed time intervals
                    now = datetime.datetime.now()
                    now_sec = now.timestamp()
                    if now_sec >= last_output + self.output_every.value:
                        self.do_output(now)
                        last_output = now_sec
                        state.check_error()
                if self.current_point.weight == 1:
                    # have added new point
                    # Callback function
                    n = self.n()
                    if n != last_n:
                        # and actually added
                        last_n = n
                        if (self.callback_function and
                                not (max(n, 1) % self.callback_every.value) and
                                self.current_point.weight == 1):
                            self.callback_function_callable(self)
                            self.last_point_callback = len(self.collection)

                        if more_than_one_process():
                            # Checking convergence and (optionally) learning
                            # the covmat of the proposal
                            if self.check_ready() and state.set(mpi.State.READY):
                                self.log.info(
                                    "%s (waiting for the rest...)", self._msg_ready)
                            if state.all_ready():
                                self.mpi_info("All chains are r%s", self._msg_ready[1:])
                                self.check_convergence_and_learn_proposal()
                                self.i_learn += 1
                        else:
                            if self.check_ready():
                                self.log.debug(self._msg_ready)
                                self.check_convergence_and_learn_proposal()
                                self.i_learn += 1
                elif self.current_point.weight % state_check_every == 0:
                    state.check_error()
                    # more frequent checks near beginning
                    state_check_every = min(10, state_check_every + 1)

            if last_n == self.max_samples:
                self.log.info("Reached maximum number of accepted steps allowed (%s). "
                              "Stopping.", self.max_samples)

            # Write the last batch of samples ( < output_every (not in sec))
            self.collection.out_update()

        ns = mpi.gather(self.n())
        self.mpi_info("Sampling complete after %d accepted steps.", sum(ns))

    def n(self, burn_in=False):
        """
        Returns the total number of accepted steps taken, including or not burn-in steps
        depending on the value of the `burn_in` keyword.
        """
        return len(self.collection) + (
            0 if not burn_in else (
                self.burn_in.value -
                self.burn_in_left // self.current_point.output_thin + 1))

    def get_new_sample_metropolis(self):
        """
        Draws a new trial point from the proposal pdf and checks whether it is accepted:
        if it is accepted, it saves the old one into the collection and sets the new one
        as the current state; if it is rejected increases the weight of the current state
        by 1.

        Returns
        -------
        ``True`` for an accepted step, ``False`` for a rejected one.
        """
        trial = self.current_point.values.copy()
        self.proposer.get_proposal(trial)
        trial_results = self.model.logposterior(trial)
        accept = self.metropolis_accept(trial_results.logpost, self.current_point.logpost)
        self.process_accept_or_reject(accept, trial, trial_results)
        return accept

    def get_new_sample_dragging(self):
        """
        Draws a new trial point in the slow subspace, and gets the corresponding trial
        in the fast subspace by "dragging" the fast parameters.
        Finally, checks the acceptance of the total step using the "dragging" pdf:
        if it is accepted, it saves the old one into the collection and sets the new one
        as the current state; if it is rejected increases the weight of the current state
        by 1.

        Returns
        -------
        ``True`` for an accepted step, ``False`` for a rejected one.
        """
        # Prepare starting and ending points *in the SLOW subspace*
        # "start_" and "end_" mean here the extremes in the SLOW subspace
        current_start_point = self.current_point.values
        current_start_logpost = self.current_point.logpost
        current_end_point = current_start_point.copy()
        self.proposer.get_proposal_slow(current_end_point)
        self.log.debug("Proposed slow end-point: %r", current_end_point)
        # Save derived parameters of delta_slow jump, in case I reject all the dragging
        # steps but accept the move in the slow direction only
        current_end = self.model.logposterior(current_end_point)
        if current_end.logpost == -np.inf:
            self.current_point.weight += 1
            return False
        # accumulators for the "dragging" probabilities to be metropolis-tested
        # at the end of the interpolation
        start_drag_logpost_acc = current_start_logpost
        end_drag_logpost_acc = current_end.logpost
        # don't compute derived during drag, unless must be computed anyway
        derived = self.model.requires_derived

        # alloc mem
        delta_fast = np.empty(len(current_start_point))
        # start dragging
        for i_step in range(1, 1 + self.drag_interp_steps):
            self.log.debug("Dragging step: %d", i_step)
            # take a step in the fast direction in both slow extremes
            delta_fast[:] = 0.
            self.proposer.get_proposal_fast(delta_fast)
            self.log.debug("Proposed fast step delta: %r", delta_fast)
            proposal_start_point = current_start_point + delta_fast
            # get the new extremes for the interpolated probability
            # (reject if any of them = -inf; avoid evaluating both if just one fails)
            # Force the computation of the (slow blocks) derived params at the starting
            # point, but discard them, since they contain the starting point's fast ones,
            # not used later -- save the end point's ones.
            proposal_start_logpost = self.model.logposterior(
                proposal_start_point, return_derived=bool(derived),
                _no_check=True).logpost
            if proposal_start_logpost != -np.inf:
                proposal_end_point = current_end_point + delta_fast
                proposal_end = self.model.logposterior(
                    proposal_end_point, return_derived=bool(derived), _no_check=True)
                if proposal_end.logpost != -np.inf:
                    # create the interpolated probability and do a Metropolis test
                    frac = i_step / (1 + self.drag_interp_steps)
                    proposal_interp_logpost = ((1 - frac) * proposal_start_logpost +
                                               frac * proposal_end.logpost)
                    current_interp_logpost = ((1 - frac) * current_start_logpost +
                                              frac * current_end.logpost)
                    accept_drag = self.metropolis_accept(proposal_interp_logpost,
                                                         current_interp_logpost)
                    if accept_drag:
                        # If the dragging step was accepted, do the drag
                        current_start_point = proposal_start_point
                        current_start_logpost = proposal_start_logpost
                        current_end_point = proposal_end_point
                        current_end = proposal_end
                else:
                    accept_drag = False
            else:
                accept_drag = False
            self.log.debug("Dragging step: %s",
                           ("accepted" if accept_drag else "rejected"))

            # In any case, update the dragging probability for the final metropolis test
            start_drag_logpost_acc += current_start_logpost
            end_drag_logpost_acc += current_end.logpost
        # Test for the TOTAL step
        n_average = 1 + self.drag_interp_steps
        accept = self.metropolis_accept(end_drag_logpost_acc / n_average,
                                        start_drag_logpost_acc / n_average)
        if accept and not derived:
            # recompute with derived parameters (slow parameter ones should be cached)
            current_end = self.model.logposterior(current_end_point)

        self.process_accept_or_reject(accept, current_end_point, current_end)
        self.log.debug("TOTAL step: %s", ("accepted" if accept else "rejected"))
        return accept

    def metropolis_accept(self, logp_trial, logp_current):
        """
        Symmetric-proposal Metropolis-Hastings test.

        Returns
        -------
        ``True`` or ``False``.
        """
        if logp_trial == -np.inf:
            return False
        if logp_trial > logp_current:
            return True
        posterior_ratio = (logp_current - logp_trial) / self.temperature
        return self._rng.standard_exponential() > posterior_ratio

    def process_accept_or_reject(self, accept_state: bool, trial: np.ndarray,
                                 trial_results: LogPosterior):
        """Processes the acceptance/rejection of the new point."""
        if accept_state:
            # add the old point to the collection (if not burning or initial point)
            if self.burn_in_left <= 0:
                if self.current_point.add_to_collection(self.collection):
                    self.log.debug("New sample, #%d: \n   %s",
                                   self.n(), self.current_point)
                    # Update chain files, if output_every *not* in sec
                    if not self.output_every.unit:
                        if self.n() % self.output_every.value == 0:
                            self.collection.out_update()
            else:
                self.burn_in_left -= 1
                self.log.debug("Burn-in sample:\n   %s", self.current_point)
                if self.burn_in_left == 0 and self.burn_in:
                    self.log.info("Finished burn-in phase: discarded %d accepted steps.",
                                  self.burn_in.value)
            # set the new point as the current one, with weight one
            self.current_point.add(trial, trial_results)
            self._prior_rejections = 0
        else:  # not accepted
            self.current_point.weight += 1
            if trial_results.logprior == -np.inf:
                self._prior_rejections += 1
            # Failure criterion: chain stuck! (but be more permissive during burn_in)
            # only stop if not a pure-prior rejection
            max_tries_now = self.max_tries.value * (
                    1 + (10 - 1) * np.sign(self.burn_in_left))
            if self.current_point.weight - self._prior_rejections > max_tries_now:
                self.collection.out_update()
                raise LoggedError(
                    self.log,
                    "The chain has been stuck for %d attempts, stopping sampling. "
                    "Make sure the reference point is sensible and initial covmat. "
                    "For parameters not included in an initial covmat, the 'proposal' "
                    "width set for each parameter should be of order of the expected "
                    "conditional posterior width, which may be much smaller than the "
                    "marginalized posterior width - choose a smaller "
                    "rather than larger value if in doubt. You can also decrease the "
                    "'proposal_scale' option for mcmc, though small values will sample "
                    "less efficiently once things converge. Or make your starting 'ref'"
                    "tighter around an expected best-fit value\n"
                    "Alternatively (though not advisable) make 'max_tries: np.inf' "
                    "(or 'max_tries: .inf' in yaml).\n"
                    "Current point: %s\nCurrent result: %s\n"
                    "Last proposal: %s\nWith rejected result: %s",
                    max_tries_now, self.current_point, self.current_point.results,
                    trial, trial_results)
            elif self.current_point.weight > max_tries_now and \
                    not getattr(self, '_prior_tries_warning'):
                self.log.warning("Proposal has been rejected %s times", max_tries_now)
                self._prior_tries_warning = True

    # Functions to check convergence and learn the covariance of the proposal distribution

    def check_ready(self):
        """
        Checks if the chain(s) is(/are) ready to check convergence and, if requested,
        learn a new covariance matrix for the proposal distribution.
        """
        n = len(self.collection)
        # If *just* (weight==1) got ready to check+learn
        if not (n % self.learn_every.value) and n > 0:
            self.log.info("Learn + convergence test @ %d samples accepted.", n)
            self.model.dump_timing()
            if more_than_one_process():
                self.been_waiting += 1
                if self.been_waiting > self.max_waiting:
                    raise LoggedError(
                        self.log, "Waiting for too long for all chains to be ready. "
                                  "Maybe one of them is stuck or died unexpectedly?")
            return True
        return False

    # noinspection PyUnboundLocalVariable
    @np.errstate(all="ignore")
    def check_convergence_and_learn_proposal(self):
        """
        Checks the convergence of the sampling process, and, if requested,
        learns a new covariance matrix for the proposal distribution from the covariance
        of the last samples.
        """
        # Compute Rminus1 of means
        self.been_waiting = 0
        if more_than_one_process():
            # Compute and gather means and covs
            use_first = int(self.n() / 2)
            mean = self.collection.mean(first=use_first, tempered=True)
            cov = self.collection.cov(first=use_first, tempered=True)
            acceptance_rate = self.get_acceptance_rate(use_first)
            Ns, means, covs, acceptance_rates = mpi.array_gather(
                [self.n(), mean, cov, acceptance_rate])
        else:
            # Compute and gather means, covs and CL intervals of last m-1 chain fractions
            m = 1 + self.Rminus1_single_split
            cut = int(len(self.collection) / m)
            try:
                acceptance_rate = self.get_acceptance_rate(cut)
                Ns = np.ones(m - 1) * cut
                ranges = [(i * cut, (i + 1) * cut - 1) for i in range(1, m)]
                means = np.array(
                    [self.collection.mean(
                        first=f, last=l, tempered=True)
                     for f, l in ranges])
                covs = np.array(
                    [self.collection.cov(
                        first=f, last=l, tempered=True)
                     for f, l in ranges])
            except always_stop_exceptions:
                raise
            except Exception:  # pylint: disable=broad-except
                self.log.info("Not enough points in chain to check convergence. "
                              "Waiting for next checkpoint.")
                return
            acceptance_rates = None
        if is_main_process():
            self.progress.at[self.i_learn, "N"] = sum(Ns)
            self.progress.at[self.i_learn, "timestamp"] = \
                datetime.datetime.now().isoformat()
            acceptance_rate = (np.average(acceptance_rates, weights=Ns)
                               if acceptance_rates is not None else acceptance_rate)
            accpt_multi_str = (" = avg(%r)" % list(acceptance_rates)
                               if acceptance_rates is not None else "")
            self.log.info(" - Acceptance rate: %.3f%s", acceptance_rate, accpt_multi_str)
            self.progress.at[self.i_learn, "acceptance_rate"] = acceptance_rate
            # "Within" or "W" term -- our "units" for assessing convergence
            # and our prospective new covariance matrix
            mean_of_covs = np.average(covs, weights=Ns, axis=0)
            # "Between" or "B" term
            # We don't weight with the number of samples in the chains here:
            # shorter chains will likely be outliers, and we want to notice them
            cov_of_means = np.atleast_2d(np.cov(means.T))  # , fweights=Ns)
            # For numerical stability, we turn mean_of_covs into correlation matrix:
            #   rho = (diag(Sigma))^(-1/2) * Sigma * (diag(Sigma))^(-1/2)
            # and apply the same transformation to the mean of covs (same eigenvals!)
            d = np.sqrt(np.diag(cov_of_means))
            corr_of_means = (cov_of_means / d).T / d
            norm_mean_of_covs = (mean_of_covs / d).T / d
            success_means = False
            converged_means = False
            # Cholesky of (normalized) mean of covs and eigvals of Linv*cov_of_means*L
            try:
                L = np.linalg.cholesky(norm_mean_of_covs)
            except np.linalg.LinAlgError:
                self.log.warning(
                    "Negative covariance eigenvectors. "
                    "This may mean that the covariance of the samples does not "
                    "contain enough information at this point. "
                    "Skipping learning a new covmat for now.")
            else:
                Linv = np.linalg.inv(L)
                try:
                    eigvals = np.linalg.eigvalsh(Linv.dot(corr_of_means).dot(Linv.T))
                    success_means = True
                except np.linalg.LinAlgError:
                    self.log.warning("Could not compute eigenvalues. "
                                     "Skipping learning a new covmat for now.")
                else:
                    Rminus1 = max(np.abs(eigvals))
                    self.progress.at[self.i_learn, "Rminus1"] = Rminus1
                    # For real square matrices, a possible def of the cond number is:
                    condition_number = Rminus1 / min(np.abs(eigvals))
                    self.log.debug(" - Condition number = %g", condition_number)
                    self.log.debug(" - Eigenvalues = %r", eigvals)
                    accpt_multi_str = \
                        " = sum(%r)" % list(Ns) if more_than_one_process() else ""
                    self.log.info(
                        " - Convergence of means: R-1 = %f after %d accepted steps%s",
                        Rminus1, sum(Ns), accpt_multi_str)
                    # Have we converged in means?
                    # (criterion must be fulfilled twice in a row)
                    converged_means = max(Rminus1, self.Rminus1_last) < self.Rminus1_stop
        else:
            mean_of_covs = None
            success_means = None
            converged_means = False
            Rminus1 = None
        success_means, converged_means = mpi.share((success_means, converged_means))
        # Check the convergence of the bounds of the confidence intervals
        # Same as R-1, but with the rms deviation from the mean bound
        # in units of the mean standard deviation of the chains
        if converged_means:
            if more_than_one_process():
                # pylint: disable=protected-access
                mcsamples = self.collection._sampled_to_getdist(
                    first=use_first, tempered=True)
                try:
                    bound = np.array([[
                        mcsamples.confidence(i, limfrac=self.Rminus1_cl_level / 2.,
                                             upper=which)
                        for i in range(self.model.prior.d())]
                        for which in [False, True]]).T
                    success_bounds = True
                except Exception:  # pylint: disable=broad-except
                    bound = None
                    success_bounds = False
                bounds = np.array(mpi.gather(bound))
            else:
                try:
                    mcsamples_list = [
                        # pylint: disable=protected-access
                        self.collection._sampled_to_getdist(
                            first=i * cut, last=(i + 1) * cut - 1, tempered=True)
                        for i in range(1, m)]
                except always_stop_exceptions:
                    raise
                except Exception:  # pylint: disable=broad-except
                    self.log.info("Not enough points in chain to check c.l. convergence. "
                                  "Waiting for next checkpoint.")
                    return
                try:
                    bounds = [
                        np.array(
                            [[mcs.confidence(
                                i, limfrac=self.Rminus1_cl_level / 2., upper=which)
                                for i in range(self.model.prior.d())]
                                for which in [False, True]]).T
                        for mcs in mcsamples_list]
                    success_bounds = True
                except Exception:  # pylint: disable=broad-except
                    bounds = None
                    success_bounds = False
            if is_main_process():
                if success_bounds:
                    Rminus1_cl = (np.std(bounds, axis=0).T /
                                  np.sqrt(np.diag(mean_of_covs)))
                    self.log.debug(" - normalized std's of bounds = %r", Rminus1_cl)
                    Rminus1_cl = np.max(Rminus1_cl)
                    self.progress.at[self.i_learn, "Rminus1_cl"] = Rminus1_cl
                    accpt_multi_str = \
                        " = sum(%r)" % list(Ns) if more_than_one_process() else ""
                    self.log.info(
                        " - Convergence of bounds: R-1 = %f after %d accepted steps%s",
                        Rminus1_cl,
                        sum(Ns) if more_than_one_process() else self.n(),
                        accpt_multi_str,
                    )
                    if Rminus1_cl < self.Rminus1_cl_stop:
                        self.converged = True
                        self.log.info("The run has converged!")
                        self._Ns = Ns
                else:
                    self.log.info("Computation of the bounds was not possible. "
                                  "Waiting until the next converge check.")
        # Broadcast and save the convergence status and the last R-1 of means
        if success_means:
            self.Rminus1_last, self.converged = mpi.share(
                (Rminus1, self.converged) if is_main_process() else None)
            # Do we want to learn a better proposal pdf?
            if self.learn_proposal and not self.converged:
                good_Rminus1 = (self.learn_proposal_Rminus1_max >
                                self.Rminus1_last > self.learn_proposal_Rminus1_min)
                if not good_Rminus1:
                    self.mpi_info("Convergence less than requested for updates: "
                                  "waiting until the next convergence check.")
                    return
                mean_of_covs = mpi.share(mean_of_covs)
                try:
                    self.proposer.set_covariance(mean_of_covs)  # is already tempered
                    self.mpi_info(" - Updated covariance matrix of proposal pdf.")
                    self.mpi_debug("%r", mean_of_covs)
                except Exception:  # pylint: disable=broad-except
                    self.mpi_debug("Updating covariance matrix failed unexpectedly. "
                                   "waiting until next covmat learning attempt.")
        # Save checkpoint info
        self.write_checkpoint()

    def do_output(self, date_time):
        """Writes/updates the output products of the chain."""
        self.collection.out_update()
        msg = "Progress @ %s : " % date_time.strftime("%Y-%m-%d %H:%M:%S")
        msg += "%d steps taken" % self.n_steps_raw
        if self.burn_in_left and self.burn_in:  # NB: burn_in_left = 1 even if no burn_in
            msg += " -- still burning in, %d accepted steps left." % self.burn_in_left
        else:
            msg += ", and %d accepted." % self.n()
        self.log.info(msg)

    def write_checkpoint(self):
        """Writes/updates the checkpoint file."""
        if is_main_process() and self.output:
            checkpoint_filename = self.checkpoint_filename()
            self.dump_covmat(remove_temperature_cov(
                self.proposer.get_covariance(), self.temperature))
            checkpoint_info = {"sampler": {self.get_name(): dict([
                ("converged", self.converged),
                ("Rminus1_last", self.Rminus1_last),
                ("burn_in", (self.burn_in.value  # initial: repeat burn-in if not finished
                             if not self.n() and self.burn_in_left else
                             0)),  # to avoid overweighting last point of prev. run
                ("mpi_size", get_mpi_size())])}}
            yaml_dump_file(checkpoint_filename, checkpoint_info, error_if_exists=False)
            if not self.progress.empty:
                with open(self.progress_filename(), "a",
                          encoding="utf-8") as progress_file:
                    fmts = {"N": "{:9d}".format}
                    progress_file.write(self.progress.tail(1).to_string(
                        header=False, index=False, formatters=fmts) + "\n")
            self.log.debug("Dumped checkpoint and progress info, and current covmat.")

    def converge_info_changed(self, old_info, new_info):
        """Whether convergence parameters have changed between two inputs."""
        converge_params = ['Rminus1_stop', "Rminus1_cl_stop", "Rminus1_cl_level",
                           "max_samples"]
        return any(old_info.get(p) != new_info.get(p) for p in converge_params)

    # Finally: returning the computed products ###########################################

    def samples(self,
                combined: bool = False,
                skip_samples: float = 0,
                to_getdist: bool = False,
                ) -> Union[SampleCollection, "MCSamples"]:
        """
        Returns the sample of accepted steps.

        Parameters
        ----------
        combined: bool, default: False
            If ``True`` and running more than one MPI process, the ``sample`` key of the
            returned dictionary contains a concatenated sample including all parallel
            chains concatenated, instead of the chain of the current process only. For
            this to work, this method needs to be called from all MPI processes
            simultaneously.
        skip_samples: int or float, default: 0
            Skips some amount of initial samples (if ``int``), or an initial fraction of
            them (if ``float < 1``). If concatenating (``combined=True``), skipping is
            applied before concatenation. Forces the return of a copy.
        to_getdist: bool, default: False
            If ``True``, returns sample collections as :class:'getdist.MCSamples`. If both
            this option and ``combined`` are ``True``, the latter is ignored and a
            multi-chain :class:'getdist.MCSamples` object is created out of the chains of
            all MPI processes.

        Returns
        -------
        SampleCollection, getdist.MCSamples
            The sample of accepted steps.
        """
        collection = self.collection.skip_samples(skip_samples, inplace=False)
        if to_getdist:
            collection = collection.to_getdist(model=self.model)
        if combined and more_than_one_process():
            collections = mpi.gather(collection)
            if is_main_process():
                if to_getdist:
                    collections[0].loadChains(
                        root=None,
                        files_or_samples=[c.samples for c in collections],
                        weights=[c.weights for c in collections],
                        loglikes=[c.loglikes for c in collections],
                    )
                    collection = collections[0]
                else:
                    if not skip_samples:
                        self.mpi_warning(
                            "When combining chains, it is recommended to remove some "
                            "initial fraction, e.g. 'skip_samples=0.3'"
                        )
                    for collection in collections[1:]:
                        # pylint: disable=protected-access
                        collections[0]._append(collection)
                    collection = collections[0]
            collection = mpi.share_mpi(collection)
        if self.temperature != 1 and not to_getdist:
            self.mpi_warning(
                "The MCMC chain(s) are stored with temperature != 1. "
                "Keep that in mind when operating on them, or detemper (in-place) with "
                "products()['sample'].reset_temperature()'.")
        return collection

    def products(
            self,
            combined: bool = False,
            skip_samples: float = 0,
            to_getdist: bool = False,
    ) -> dict:
        """
        Returns the products of the sampling process.

        Parameters
        ----------
        combined: bool, default: False
            If ``True`` and running more than one MPI process, the ``sample`` key of the
            returned dictionary contains a concatenated sample including all parallel
            chains concatenated, instead of the chain of the current process only. For
            this to work, this method needs to be called from all MPI processes
            simultaneously.
        skip_samples: int or float, default: 0
            Skips some amount of initial samples (if ``int``), or an initial fraction of
            them (if ``float < 1``). If concatenating (``combined=True``), skipping is
            applied previously to concatenation. Forces the return of a copy.
        to_getdist: bool, default: False
            If ``True``, returns sample collections as :class:'getdist.MCSamples`. If both
            this option and ``combined`` are ``True``, the latter is ignored and a
            multi-chain :class:'getdist.MCSamples` object is created out of the chains of
            all MPI processes.

        Returns
        -------
        dict
            A dictionary containing the :class:`cobaya.collection.SampleCollection` of
            accepted steps under ``"sample"``, and a progress report table under
            ``"progress"``.
        """
        products = {
            "sample": self.samples(
                combined=combined,
                skip_samples=skip_samples,
                to_getdist=to_getdist
            )
        }
        products["progress"] = share_mpi(getattr(self, "progress", None))
        return products

    # Class methods
    @classmethod
    def output_files_regexps(cls, output, info=None, minimal=False):
        """Returns regexps for the output files created by this sampler."""
        regexps = [output.collection_regexp(name=None)]
        if minimal:
            return [(r, None) for r in regexps]
        regexps += [
            re.compile(output.prefix_regexp_str + re.escape(ext.lstrip(".")) + "$")
            for ext in [Extension.checkpoint, Extension.progress, Extension.covmat]]
        return [(r, None) for r in regexps]

    @classmethod
    def get_version(cls):
        """
        Returns the version string of this samples (since it is built-in, that of Cobaya).
        """
        return get_version()

    @classmethod
    def _get_desc(cls, info=None):
        drag_string = r" using the fast-dragging procedure described in \cite{Neal:2005}"
        if info is None:
            # Unknown case (no info passed)
            string = " [(if drag: True)%s]" % drag_string
        else:
            string = drag_string if info.get("drag", cls.get_defaults()["drag"]) else ""
        return ("Adaptive, speed-hierarchy-aware MCMC sampler (adapted from CosmoMC) "
                r"\cite{Lewis:2002ah,Lewis:2013hha}" + string + ".")


# Plotting tool for chain progress #######################################################

def plot_progress(progress, ax=None, index=None,
                  figure_kwargs=empty_dict, legend_kwargs=empty_dict):
    """
    Plots progress of one or more MCMC runs: evolution of R-1
    (for means and c.l. intervals) and acceptance rate.

    Takes a ``progress`` instance (actually a ``pandas.DataFrame``,
    returned as part of the sampler ``products``),
    a chain ``output`` prefix, or a list of those
    for plotting progress of several chains at once.

    You can use ``figure_kwargs`` and ``legend_kwargs`` to pass arguments to
    ``matplotlib.pyplot.figure`` and ``matplotlib.pyplot.legend`` respectively.

    Returns a subplots axes array. Display with ``matplotlib.pyplot.show()``.

    """
    if ax is None:
        import matplotlib.pyplot as plt  # pylint: disable=import-outside-toplevel
        # noinspection PyTypeChecker
        fig, ax = plt.subplots(nrows=2, sharex=True, **figure_kwargs)
    if isinstance(progress, DataFrame):
        pass  # go on to plotting
    elif isinstance(progress, str):
        try:
            if not progress.endswith(Extension.progress):
                progress += Extension.progress
            progress = load_DataFrame(progress)
            # 1-based
            progress.index = np.arange(1, len(progress) + 1)
        except Exception as excpt:
            raise ValueError(
                "Cannot load progress file %r: %s" % (progress, str(excpt))) from excpt
    elif hasattr(type(progress), "__iter__"):
        # Assume is a list of progress'es
        for i, p in enumerate(progress):
            plot_progress(p, ax=ax, index=i + 1)
        return ax
    else:
        raise ValueError("Cannot understand progress argument: %r" % progress)
    # Plot!
    tag_pre = "" if index is None else "%d : " % index
    p = ax[0].semilogy(progress.N, progress.Rminus1,
                       "o-", label=tag_pre + "means")
    ax[0].semilogy(progress.N, progress.Rminus1_cl,
                   "x:", c=p[0].get_color(), label=tag_pre + "bounds")
    ax[0].set_ylabel(r"$R-1$")
    ax[0].legend(**legend_kwargs)
    ax[1].plot(progress.N, progress.acceptance_rate, "o-")
    ax[1].set_ylabel(r"acc. rate")
    return ax



"""
.. module:: install

:Synopsis: Tools and script to install the external code and data packages needed by
           the Cobaya components to be used.
:Author: Jesus Torrado

"""
# Global
import os
import sys
import re
import subprocess
import traceback
import shutil
import tempfile
import logging
from itertools import chain
from packaging import version
import importlib
import requests  # type: ignore
import tqdm  # type: ignore
from typing import List, Mapping, Union

# Local
from cobaya.log import logger_setup, LoggedError, NoLogging, get_logger
from cobaya.component import get_component_class, ComponentNotFoundError
from cobaya.tools import create_banner, warn_deprecation, \
    write_packages_path_in_config_file, get_config_path, VersionCheckError, \
    resolve_packages_path, similar_internal_class_names, find_with_regexp
from cobaya.input import get_used_components
from cobaya.conventions import code_path, data_path, packages_path_arg, \
    packages_path_env, Extension, install_skip_env, packages_path_arg_posix, \
    packages_path_config_file, packages_path_input
from cobaya.mpi import set_mpi_disabled
from cobaya.typing import InputDict, InfoDict
from cobaya.yaml import yaml_load_file

_banner_symbol = "="
_banner_length = 80
_version_filename = "version.dat"


def do_package_install(component: str, package_install: Union[InfoDict, str],
                       full_code_path: str, logger):
    # attempt to install package if package_install download info is present
    # similar to InstallableLikelihood install_options (could be refactored)
    component_root = component.split('.')[0]
    directory = package_install.get("directory")
    if package_install == "pip":
        package_install = {"pip": None}
    elif not isinstance(package_install, Mapping):
        raise LoggedError(logger, "Invalid package_install: must be 'pip' or a "
                                  "dictionary with download_url or github_repository")
    package = package_install.get("pip") or component_root
    cwd = None
    if repo := package_install.get("github_repository"):
        logger.info('Downloading code from github (%s)', repo)
        directory = directory or repo.split("/")[-1]
        if not download_github_release(
                full_code_path, repo, package_install.get("github_release"),
                directory=directory, logger=logger):
            return False
        cwd = os.path.join(full_code_path, directory)
        package = '.'

    elif url := package_install.get("download_url"):
        logger.info('Downloading code from %s', url)
        cwd = os.path.join(full_code_path, directory or component_root)
        if not download_file(url, cwd, logger=logger):
            return False
        paths = find_with_regexp('setup\\.py', cwd, True)
        if not paths:
            raise LoggedError(logger, "No setup.py found in %s for %s", cwd, component)
        cwd = os.path.dirname(paths[0])
        package = '.'

    elif "pip" not in package_install:
        raise LoggedError(logger, "Invalid package_install: must define pip, "
                                  "github_repository or download_url")

    logger.info('pip installing %s', component_root)
    return not pip_install(package, upgrade=True, logger=logger, cwd=cwd)


def install(*infos, **kwargs):
    """
    Installs the external packages required by the components mentioned in ``infos``.

    ``infos`` can be input dictionaries, single component names, or names of yaml files

    :param force: force re-installation of apparently installed packages (default:
       ``False``).
    :param test: just check whether components are installed  (default: ``False``).
    :param upgrade: force upgrade of obsolete components (default: ``False``).
    :param skip: keywords of components that will be skipped during installation.
    :param skip_global: skip installation of already-available Python modules (default:
       ``False``).
    :param debug: produce verbose debug output  (default: ``False``).
    :param code: set to ``False`` to skip code packages (default: ``True``).
    :param data: set to ``False`` to skip data packages (default: ``True``).
    :param no_progress_bars: no progress bars shown; use when output is saved into a text
       file (e.g. when running on a cluster) (default: ``False``).
    :param no_set_global: do not store the installation path for later runs (default:
       ``False``).

    :return: ``True`` if all components are installed without skips, ``False`` if
                  some components skipped, None (or exception raised) otherwise
    """
    debug = kwargs.get("debug", False)
    logger = kwargs.get("logger")
    if not logger:
        logger_setup(debug=debug)
        logger = get_logger("install")
    path = kwargs.get("path")
    infos = [yaml_load_file(info)
             if isinstance(info, str) and info.split('.')[-1] in Extension.yamls
             else info for info in infos]
    infos_not_single_names = [info for info in infos if isinstance(info, Mapping)]
    if not path:
        path = resolve_packages_path(*infos_not_single_names)
    if not path:
        raise LoggedError(
            logger, ("No 'path' argument given, and none could be found in input infos "
                     "(as %r), the %r env variable or the config file. "
                     "Maybe specify one via a command line argument '-%s [...]'?"),
            packages_path_input, packages_path_env, packages_path_arg[0])
    # General install path for all dependencies
    general_abspath = os.path.abspath(path)
    logger.info("Installing external packages at '%s'", general_abspath)
    # Set the installation path in the global config file
    if not kwargs.get("no_set_global") and not kwargs.get("test"):
        write_packages_path_in_config_file(general_abspath)
        logger.info(
            "The installation path has been written into the global config file: %s",
            os.path.join(get_config_path(), packages_path_config_file))
    kwargs_install = {"force": kwargs.get("force", False),
                      "no_progress_bars": kwargs.get("no_progress_bars"),
                      "show_error": False}
    for what in (code_path, data_path):
        kwargs_install[what] = kwargs.get(what, True)
        spath = os.path.join(general_abspath, what)
        if kwargs_install[what] and not os.path.exists(spath):
            try:
                os.makedirs(spath)
            except OSError:
                raise LoggedError(
                    logger, f"Could not create the desired installation folder '{spath}'")
    # To check e.g. for a version upgrade, it needs to reload the component class and
    # all relevant imported modules: the implementation of `is_installed` for each
    # class is expected to always reload external modules if passed `reload=True`
    # (should be False by default to avoid deleting objects unnecessarily).
    kwargs_is_installed = {"reload": True}
    unknown_components = []  # could not be identified
    failed_components = []  # general errors
    obsolete_components = []  # older or unknown version already installed
    skip_keywords_arg = set(kwargs.get("skip") or [])
    # NB: if passed with quotes as `--skip "a b"`, it's interpreted as a single key
    skip_keywords_arg = set(chain(*[word.split() for word in skip_keywords_arg]))
    skip_keywords_env = set(
        os.environ.get(install_skip_env, "").replace(",", " ").lower().split())
    skip_keywords = skip_keywords_arg.union(skip_keywords_env)
    skipped = False
    # Combine all requested components and install them
    # NB: components mentioned by name may be repeated with those given in dict infos.
    #     That's OK, because the install check will skip them in the 2nd pass
    used_components, components_infos = get_used_components(*infos, return_infos=True)
    for kind, components in used_components.items():
        for component in components:
            name_w_kind = (kind + ":" if kind else "") + component
            print()
            print(create_banner(name_w_kind,
                                symbol=_banner_symbol, length=_banner_length), end="")
            print()
            if _skip_helper(component.lower(), skip_keywords, skip_keywords_env, logger):
                skipped = True
                continue
            info = components_infos[component]
            if isinstance(info, str) or "external" in info:
                logger.info(
                    f"Component '{name_w_kind}' is a custom function. Nothing to do.")
                continue
            class_name = (info or {}).get("class")
            if class_name:
                logger.info(f"Class to be installed for this component: {class_name}")
            python_path = info.pop("python_path", None)
            package_install = info.get("package_install") or {}
            min_version = package_install.get('min_version')

            def _imported_class():
                return get_component_class(
                    component, kind=kind, component_path=python_path,
                    class_name=class_name, logger=logger, min_package_version=min_version,
                    allow_internal=not package_install)

            try:
                try:
                    if package_install and (
                            kwargs_install["force"] or kwargs.get("upgrade")):
                        raise ComponentNotFoundError(logger)
                    imported_class = _imported_class()
                except (ComponentNotFoundError, VersionCheckError):
                    if package_install:
                        if do_package_install(component, package_install,
                                              os.path.join(general_abspath, code_path),
                                              logger):
                            importlib.invalidate_caches()
                            imported_class = _imported_class()
                        else:
                            logger.error(
                                f"Package install failed for '{name_w_kind}'. Skipping.")
                            unknown_components += [name_w_kind]
                            continue
                    else:
                        raise
                except Exception:
                    raise
            except ComponentNotFoundError:
                logger.error(
                    f"Component '{name_w_kind}' could not be identified. Skipping.")
                unknown_components += [name_w_kind]
                continue
            except Exception:
                traceback.print_exception(*sys.exc_info(), file=sys.stdout)
                logger.error(f"An error occurred when loading '{name_w_kind}'. Skipping.")
                failed_components += [name_w_kind]
                continue
            else:
                if _skip_helper(imported_class.__name__.lower(), skip_keywords,
                                skip_keywords_env, logger):
                    skipped = True
                    continue
            # Update the name if the kind was unknown
            if not kind:
                name_w_kind = imported_class.get_kind() + ":" + component
            is_compatible = getattr(imported_class, "is_compatible", lambda: True)()
            if not is_compatible:
                logger.error(f"Skipping '{name_w_kind}' "
                             "because it is not compatible with your OS.")
                failed_components += [name_w_kind]
                continue
            logger.info("Checking if dependencies have already been installed...")
            is_installed = getattr(imported_class, "is_installed", None)
            if is_installed is None:
                logger.info(f"Component '{name_w_kind}' is a fully built-in component: "
                            "nothing to do.")
                continue
            this_component_install_path = general_abspath
            get_path = getattr(imported_class, "get_path", None)
            if get_path:
                this_component_install_path = get_path(this_component_install_path)
            # Check previous installations and their versions
            has_been_installed = False
            is_old_version_msg = None
            with NoLogging(None if debug else logging.ERROR):
                try:
                    if kwargs.get("skip_global"):
                        has_been_installed = is_installed(
                            path="global", **kwargs_install, **kwargs_is_installed)
                    if not has_been_installed:
                        has_been_installed = is_installed(
                            path=this_component_install_path, **kwargs_install,
                            **kwargs_is_installed)
                except VersionCheckError as excpt:
                    is_old_version_msg = str(excpt)
            if has_been_installed:  # no VersionCheckError was raised
                logger.info("External dependencies for this component already installed.")
                if kwargs.get("test"):
                    continue
                if kwargs_install["force"] and not kwargs.get("skip_global"):
                    logger.info("Forcing re-installation, as requested.")
                else:
                    logger.info("Doing nothing.")
                    continue
            elif is_old_version_msg:
                logger.info(f"Version check failed: {is_old_version_msg}")
                obsolete_components += [name_w_kind]
                if kwargs.get("test"):
                    continue
                if not kwargs.get("upgrade") and not kwargs.get("force"):
                    logger.info("Skipping because '--upgrade' not requested.")
                    continue
            else:
                logger.info("Check found no existing installation")
                if not debug:
                    logger.info(
                        "(If you expected this to be already installed, re-run "
                        "`cobaya-install` with --debug to get more verbose output.)")
                if kwargs.get("test"):
                    # We are only testing whether it was installed, so consider it failed
                    failed_components += [name_w_kind]
                    continue
            # Do the install
            logger.info("Installing...")
            install_args = dict(kwargs_install, show_error=True)
            try:
                install_this = getattr(imported_class, "install", None)
                success = install_this(path=general_abspath, **install_args)
            except Exception:
                traceback.print_exception(*sys.exc_info(), file=sys.stdout)
                logger.error("An unknown error occurred. Delete the external packages "
                             "folder %r and try again. "
                             "Please, notify the developers if this error persists.",
                             general_abspath)
                success = False
            if success:
                logger.info("Successfully installed! Let's check it...")
            else:
                logger.error(
                    "Installation failed! Look at the error messages above. "
                    "Solve them and try again, or, if you are unable to solve them, "
                    "install the packages required by this component manually.")
                failed_components += [name_w_kind]
                continue
            # Test installation
            reloaded_class = get_component_class(
                component, kind=kind, component_path=info.pop("python_path", None),
                class_name=class_name, logger=logger)
            reloaded_is_installed = getattr(reloaded_class, "is_installed", None)
            with NoLogging(None if debug else logging.ERROR):
                try:
                    successfully_installed = reloaded_is_installed(
                        path=this_component_install_path, **kwargs_install,
                        **kwargs_is_installed)
                except Exception:
                    traceback.print_exception(*sys.exc_info(), file=sys.stdout)
                    successfully_installed = False
            if not successfully_installed:
                logger.error("Installation apparently worked, "
                             "but the subsequent installation test failed! "
                             "This does not always mean that there was an actual error, "
                             "and is sometimes fixed simply by running the installer "
                             "again. If not, look closely at the error messages above, "
                             "or re-run with --debug for more verbose output. "
                             "If you are unable to fix the issues above, "
                             "try installing the packages required by this "
                             "component manually.")
                failed_components += [name_w_kind]
            else:
                logger.info("Installation check successful.")
    print()
    print(create_banner(" * Summary * ",
                        symbol=_banner_symbol, length=_banner_length), end="")
    print()
    bullet = "\n - "
    if unknown_components:
        suggestions_dict = {
            name: similar_internal_class_names(name) for name in unknown_components}
        suggestions_msg = \
            bullet + bullet.join(
                f"{name}: did you mean any of the following? {sugg} "
                "(mind capitalization!)" for name, sugg in suggestions_dict.items())
        raise LoggedError(
            logger, ("The following components could not be identified and were skipped:"
                     f"{suggestions_msg}"))
    if failed_components:
        raise LoggedError(
            logger, ("The installation (or installation test) of some component(s) has "
                     "failed: %s\nCheck output of the installer of each component above "
                     "for precise error info.\n"),
            bullet + bullet.join(failed_components))
    if obsolete_components:
        raise LoggedError(
            logger, ("The following packages are obsolete. Re-run with `--upgrade` option"
                     " (not upgrading by default to preserve possible user changes): %s"),
            bullet + bullet.join(obsolete_components))
    if not unknown_components and not failed_components and not obsolete_components:
        logger.info("All requested components' dependencies correctly installed at "
                    f"{general_abspath}")
        return not skipped


def _skip_helper(name, skip_keywords, skip_keywords_env, logger):
    try:
        this_skip_keyword = next(s for s in skip_keywords
                                 if s.lower() in name.lower())
        env_msg = (" in env var %r" % install_skip_env
                   if this_skip_keyword in skip_keywords_env else "")
        logger.info("Skipping %r as per skip keyword %r" + env_msg,
                    name, this_skip_keyword)
        return True
    except StopIteration:
        return False


def download_file(url, path, *, size=None, no_progress_bars=False, logger=None):
    """
    Downloads (and optionally decompresses) a file into a given path.

    :param url: url from which to download the file.
    :param path: path to where the file should be downloaded.
    :param size: size in bytes of the file to download; can be used to show percentages
       when the url headers do not show the actual size.
    :param no_progress_bars: no progress bars shown; use when output is saved into a text
       file (e.g. when running on a cluster).
    :param logger: logger to use for reporting information; a new logger is created if not
       specified.
    :return: ``True`` if the download (and decompression, if requested) was successfull,
       and ``False`` otherwise.
    """
    logger = logger or get_logger("install")
    with tempfile.TemporaryDirectory() as tmp_path:
        try:
            req = requests.get(url, allow_redirects=True, stream=True)
            # get hinted filename if available:
            try:
                filename = re.findall(
                    "filename=(.+)", req.headers['content-disposition'])[0]
                filename = filename.strip('"\'')
            except KeyError:
                filename = os.path.basename(url)
            filename_tmp_path = os.path.normpath(os.path.join(tmp_path, filename))
            size = size or int(req.headers.get('content-length', 0))
            # Adapted from https://gist.github.com/yanqd0/c13ed29e29432e3cf3e7c38467f42f51
            if not no_progress_bars:
                bar = tqdm.tqdm(total=size, unit='iB', unit_scale=True, unit_divisor=1024)
            with open(filename_tmp_path, 'wb') as f:
                for data in req.iter_content(chunk_size=1024):
                    chunk_size = f.write(data)
                    if not no_progress_bars:
                        bar.update(chunk_size)
            if not no_progress_bars:
                bar.close()
            if os.path.getsize(filename_tmp_path) < 1024:  # 1kb
                with open(filename_tmp_path, "r") as f:
                    lines = f.readlines()
                    if lines[0].startswith("404") or "not found" in lines[0].lower():
                        raise ValueError("File not found (404)!")
            logger.info('Downloaded filename %s', filename)
        except Exception as excpt:
            logger.error(
                "Error downloading %s' to folder '%s': %s", url, tmp_path, excpt)
            return False
        logger.debug('Got: %s', filename)
        extension = os.path.splitext(filename)[-1][1:]
        if not os.path.exists(path):
            os.makedirs(path)
        try:
            if extension == "zip":
                from zipfile import ZipFile
                with ZipFile(filename_tmp_path, 'r') as zipObj:
                    zipObj.extractall(path)
            else:
                import tarfile
                if extension == "tgz":
                    extension = "gz"
                with tarfile.open(filename_tmp_path, "r:" + extension) as tar:
                    tar.extractall(path)
            logger.debug('Decompressed: %s', filename)
            return True
        except Exception as excpt:
            logger.error("Error decompressing downloaded file! Corrupt file? [%s]", excpt)
            return False


def download_github_release(base_directory, repo_name, release_name=None, *, asset=None,
                            directory=None, no_progress_bars=False, logger=None):
    """
    Downloads a release (i.e. a tagged commit) or a release asset from a GitHub repo.

    :param base_directory: directory into which the release will be downloaded; will be
       created if it does not exist.
    :param repo_name: repository name as ``user/repo``.
    :param release_name: name or tag of the release.
                         If not set, uses latest default branch.
    :param asset: download just an asset (attached file) from a release.
    :param directory: name of the directory that will contain the asset or release, if
       different ``repo_name``.
    :param no_progress_bars: no progress bars shown; use when output is saved into a text
       file (e.g. when running on a cluster).
    :param logger: logger to use for reporting information; a new logger is created if not
       specified.
    :return: ``True`` if the download was successfull, and ``False`` otherwise.
    """
    logger = logger or get_logger("install")
    if "/" in repo_name:
        github_user = repo_name[:repo_name.find("/")]
        repo_name = repo_name[repo_name.find("/") + 1:]
    else:
        github_user = "CobayaSampler"
    base_url = r"https://github.com/" + github_user + "/" + repo_name
    download_directory = base_directory
    if not release_name:
        data = requests.get(
            "https://api.github.com/repos/" + github_user + "/" + repo_name).json()
        release_name = (data or {}).get('default_branch', 'main')
    if asset:
        url = (base_url + "/releases/download/" + release_name + "/" + asset)
        # Assest would get decompressed in base directory
        download_directory = os.path.join(download_directory, directory or repo_name)
    else:
        url = (base_url + "/archive/" + release_name + ".tar.gz")

    if not download_file(url, download_directory,
                         no_progress_bars=no_progress_bars, logger=logger):
        return False
    # In releases, not assets, remove version number from directory name
    # and rename if requested
    if not asset:
        w_version = next(d for d in os.listdir(base_directory)
                         if (d.startswith(repo_name) and len(d) != len(repo_name)))
        actual_download_directory = os.path.join(base_directory, w_version)
        download_directory = os.path.join(base_directory, directory or repo_name)
        if os.path.exists(download_directory):
            shutil.rmtree(download_directory)
        os.rename(actual_download_directory, download_directory)
    # Now save the version into a file to be checked later
    with open(os.path.join(download_directory, _version_filename), "w") as f:
        f.write(release_name)
    logger.info((f"{asset} from " if asset else "") +
                "%s %s downloaded and decompressed correctly.", repo_name, release_name)
    return True


def pip_install(packages, upgrade=False, logger=None, options=(), **kwargs):
    """
    Takes package name or list of them.

    Returns exit status.
    """
    if hasattr(packages, "split"):
        packages = [packages]
    cmd = [sys.executable, '-m', 'pip', 'install']
    if upgrade:
        cmd += ['--upgrade']
    cmd += list(options)
    res = subprocess.call(cmd + packages, **kwargs)
    if res:
        msg = f"pip: error installing packages '{packages}'"
        logger.error(msg) if logger else print(msg, file=sys.stderr)
    return res


def check_gcc_version(min_version="6.4", error_returns=None):
    """
    Checks the version of the ``gcc`` compiler installed.

    If the installed version is higher than ``min_version``, return ``True``,
    and ``False`` otherwise.

    If an error is produced, returns ``error_returns``.
    """
    try:
        ver = subprocess.check_output(
            "gcc -dumpversion", shell=True, stderr=subprocess.STDOUT).decode().strip()
    except subprocess.CalledProcessError:
        return error_returns
    # Change in gcc >= 7: -dumpversion only dumps major version
    if "." not in ver:
        ver = subprocess.check_output(
            "gcc -dumpfullversion", shell=True, stderr=subprocess.STDOUT).decode().strip()
    return version.parse(str(min_version)) <= version.parse(ver)


# Command-line script ####################################################################

def install_script(args=None):
    """Command line script for the installer."""
    set_mpi_disabled()
    warn_deprecation()
    # Parse arguments
    import argparse
    parser = argparse.ArgumentParser(
        prog="cobaya install",
        description="Cobaya's installation tool for external packages.")
    parser.add_argument("files_or_components", action="store", nargs="+",
                        metavar="input_file.yaml|component_name",
                        help="One or more input files or component names "
                             "(or simply 'cosmo' to install all the requisites for basic"
                             " cosmological runs)")
    parser.add_argument("-" + packages_path_arg[0], "--" + packages_path_arg_posix,
                        action="store", required=False,
                        metavar="/packages/path", default=None,
                        help="Desired path where to install external packages. "
                             "Optional if one has been set globally or as an env variable"
                             " (run with '--show_%s' to check)." %
                             packages_path_arg_posix)
    output_show_packages_path = resolve_packages_path()
    if output_show_packages_path and os.environ.get(packages_path_env):
        output_show_packages_path += " (from env variable %r)" % packages_path_env
    elif output_show_packages_path:
        output_show_packages_path += " (from config file)"
    else:
        output_show_packages_path = "(Not currently set.)"
    parser.add_argument("--show-" + packages_path_arg_posix, action="version",
                        version=output_show_packages_path,
                        help="Prints default external packages installation folder "
                             "and exits.")
    parser.add_argument("-" + "f", "--" + "force", action="store_true", default=False,
                        help="Force re-installation of apparently installed packages.")
    parser.add_argument("--%s" % "test", action="store_true", default=False,
                        help="Just check whether components are installed.")
    parser.add_argument("--upgrade", action="store_true", default=False,
                        help="Force upgrade of obsolete components.")
    parser.add_argument("--skip", action="store", nargs="*",
                        metavar="keyword", help=("Keywords of components that will be "
                                                 "skipped during installation."))
    parser.add_argument("--skip-global", action="store_true", default=False,
                        help="Skip installation of already-available Python modules.")
    parser.add_argument("-" + "d", "--" + "debug", action="store_true",
                        help="Produce verbose debug output.")
    group_just = parser.add_mutually_exclusive_group(required=False)
    group_just.add_argument("-C", "--just-code", action="store_false", default=True,
                            help="Install code of the components.", dest=data_path)
    group_just.add_argument("-D", "--just-data", action="store_false", default=True,
                            help="Install data of the components.", dest=code_path)
    parser.add_argument("--no-progress-bars", action="store_true", default=False,
                        help=("No progress bars shown; use when output is saved into a "
                              "text file (e.g. when running on a cluster)."))
    parser.add_argument("--no-set-global", action="store_true", default=False,
                        help="Do not store the installation path for later runs.")
    arguments = parser.parse_args(args)
    # Configure the logger ASAP
    logger_setup(arguments.debug)
    logger = get_logger("install")
    # Gather requests
    infos: List[Union[InputDict, str]] = []
    for f in arguments.files_or_components:
        if f.lower() == "cosmo":
            logger.info("Installing basic cosmological packages.")
            from cobaya.cosmo_input import install_basic
            infos += [install_basic]
        elif f.lower() == "cosmo-tests":
            logger.info("Installing *tested* cosmological packages.")
            from cobaya.cosmo_input import install_tests
            infos += [install_tests]
        elif os.path.splitext(f)[1].lower() in Extension.yamls:
            from cobaya.input import load_input
            infos += [load_input(f)]
        else:  # a single component name, no kind specified
            infos += [f]
    # Launch installer
    install(*infos, path=getattr(arguments, packages_path_arg), logger=logger,
            **{arg: getattr(arguments, arg)
               for arg in ["force", code_path, data_path, "no_progress_bars", "test",
                           "no_set_global", "skip", "skip_global", "debug", "upgrade"]})


if __name__ == '__main__':
    install_script()



"""
.. module:: planck_pliklite

:Synopsis: Definition of python-native nuisance-free CMB likelihoods: e.g. plik_lite
:Author: Erminia Calabrese, Antony Lewis

Nuisance-marginalized likelihood, based on covarianced and binned CL, with settings read
from .dataset file.

"""

# Global
import os
import numpy as np

# Local
from cobaya.likelihoods.base_classes import DataSetLikelihood

cl_names = ['tt', 'te', 'ee']


class PlanckPlikLite(DataSetLikelihood):
    install_options = {"github_repository": "CobayaSampler/planck_native_data",
                       "github_release": "v1", "asset": "plik_lite_2018_AL.zip",
                       "directory": "planck_2018_pliklite_native"}

    bibtex_file = 'planck2018.bibtex'
    type = "CMB"

    def init_params(self, ini):
        self.use_cl = [c.lower() for c in ini.list('use_cl')]
        assert len(self.use_cl)
        nbintt = ini.int('nbintt')
        nbinte = ini.int('nbinte')
        nbinee = ini.int('nbinee')
        lmax = ini.int('lmax')
        use_bins = ini.int_list('use_bins', [])
        bins_for_L_range = ini.int_list('bins_for_L_range', [])
        self.calibration_param = ini.string('calibration_param', 'A_planck')
        assert len(bins_for_L_range) in [0, 2]

        data = np.loadtxt(ini.relativeFileName('data'))

        bin_lmin_offset = ini.int('bin_lmin_offset')
        self.blmin = np.loadtxt(ini.relativeFileName('blmin')).astype(
            int) + bin_lmin_offset
        self.blmax = np.loadtxt(ini.relativeFileName('blmax')).astype(
            int) + bin_lmin_offset
        self.lav = (self.blmin + self.blmax) // 2
        weights = np.loadtxt(ini.relativeFileName('weights'))
        ls = np.arange(len(weights)) + bin_lmin_offset
        self.bin_lmin_offset = bin_lmin_offset
        weights *= 2 * np.pi / ls / (ls + 1)  # we work directly with  DL not CL
        self.weights = np.hstack((np.zeros(bin_lmin_offset), weights))

        self.nbins = nbintt + nbinee + nbinte

        bin_cov_file = ini.relativeFileName('cov_file_binary')

        if os.path.exists(bin_cov_file):
            from scipy.io import FortranFile
            f = FortranFile(bin_cov_file, 'r')
            cov = f.read_reals(dtype=float).reshape((self.nbins, self.nbins))
            cov = np.tril(cov) + np.tril(cov, -1).T  # make symmetric
        else:
            cov = np.loadtxt(ini.relativeFileName('cov_file'))
            # full n row x n col matrix converted from fortran binary

        self.lmax = lmax

        maxbin = max(nbintt, nbinte, nbinee)
        assert (cov.shape[0] == self.nbins)
        self.lav = self.lav[:maxbin]

        if len(use_bins) and np.max(use_bins) >= maxbin:
            raise ValueError('use_bins has bin index out of range')
        if len(bins_for_L_range):
            if len(use_bins):
                raise ValueError('can only use one bin filter')
            use_bins = [use_bin for use_bin in range(maxbin)
                        if
                        bins_for_L_range[0] <= (
                                self.blmin[use_bin] + self.blmax[use_bin]) / 2 <=
                        bins_for_L_range[1]]
            print('Actual L range: %s - %s' % (
                self.blmin[use_bins[0]], self.blmax[use_bins[-1]]))

        self.used = np.zeros(3, dtype=bool)
        self.used_bins = []
        used_indices = []
        offset = 0
        self.bandpowers = {}
        self.errors = {}

        for i, (cl, nbin) in enumerate(zip(cl_names, [nbintt, nbinte, nbinee])):
            self.used[i] = cl_names[i] in self.use_cl
            sc = self.lav[:nbin] * (self.lav[:nbin] + 1) / 2. / np.pi
            self.bandpowers[cl] = data[offset:offset + nbin, 1] * sc
            self.errors[cl] = data[offset:offset + nbin, 2] * sc
            if self.used[i]:
                if len(use_bins):
                    self.used_bins.append(
                        np.array([use_bin for use_bin in use_bins if use_bin < nbin],
                                 dtype=int))
                else:
                    self.used_bins.append(np.arange(nbin, dtype=int))
                used_indices.append(self.used_bins[-1] + offset)
            else:
                self.used_bins.append(np.arange(0, dtype=int))
            offset += nbin
        self.used_indices = np.hstack(used_indices)
        assert (self.nbins == cov.shape[0] == data.shape[0])
        self.X_data = data[self.used_indices, 1]
        self.cov = cov[np.ix_(self.used_indices, self.used_indices)]
        self.invcov = np.linalg.inv(self.cov)

    def get_requirements(self):
        # State requisites to the theory code
        self.l_max = self.lmax
        return {"Cl": {cl: self.l_max for cl in self.use_cl}}

    def binning_matrix(self, ix=0):
        # not used by main likelihood code
        lmax = self.blmax[self.used_bins[ix][-1]]
        lmin = self.blmin[self.used_bins[ix][0]]
        m = np.zeros((len(self.used_bins[ix]), lmax - lmin + 1))
        for i in self.used_bins[ix]:
            m[i, self.blmin[i] - lmin:self.blmax[i] + 1 - lmin] = self.weights[
                                                                  self.blmin[i]:
                                                                  self.blmax[i] + 1]
        return lmin, lmax, m

    def get_chi_squared(self, L0, ctt, cte, cee, A_planck=1):
        cl = np.empty(self.used_indices.shape)
        ix = 0
        for tp, cell in enumerate([ctt, cte, cee]):
            for i in self.used_bins[tp]:
                cl[ix] = np.dot(cell[self.blmin[i] - L0:self.blmax[i] - L0 + 1],
                                self.weights[self.blmin[i]:self.blmax[i] + 1])
                ix += 1
        cl /= A_planck ** 2
        diff = self.X_data - cl
        return self._fast_chi_squared(self.invcov, diff)

    def chi_squared(self, c_l_arr, A_planck=1):
        r"""
        Get chi squared from CL array from file

        :param c_l_arr: file of L and L(L+1)CL/2\pi values for C_TT, C_TE, C_EE
        :param A_planck: calibration parameter
        :return: chi-squared
        """
        L0 = int(c_l_arr[0, 0])
        return self.get_chi_squared(L0, c_l_arr[:, 1], c_l_arr[:, 2], c_l_arr[:, 3],
                                    A_planck)

    def logp(self, **data_params):
        Cls = self.provider.get_Cl(ell_factor=True)
        return -0.5 * self.get_chi_squared(0, Cls.get('tt'), Cls.get('te'), Cls.get('ee'),
                                           data_params[self.calibration_param])



# Notes about writing more presets:
# ---------------------------------
# - all parameter names below are PLANCK parameter names. They are substituted by the
#   theory-code-specific ones in `create_input`
# - don't use extra_args for precision parameters! because if the same precision param
#   is mentioned twice at the same time in different fields with different values, there
#   is no facility to take the max (or min). Instead, codify precision needs in terms of
#   requirements in the .must_provide method of the cosmo code.

# Global
from copy import deepcopy
from cobaya.typing import InfoDict

none = "(None)"
error_msg = "error_msg"

# Theory codes
theory: InfoDict = {"camb": None, "classy": None}

# Primordial perturbations
primordial = {
    'SFSR': {'desc': 'Adiabatic scalar perturbations, power law spectrum',
             'theory': theory,
             'params': {
                 'logA': {'prior': {'min': 1.61, 'max': 3.91},
                          'ref': {'dist': 'norm', 'loc': 3.05, 'scale': 0.001},
                          'proposal': 0.001, 'latex': '\\log(10^{10} A_\\mathrm{s})',
                          'drop': True},
                 'As': {'value': 'lambda logA: 1e-10*np.exp(logA)',
                        'latex': 'A_\\mathrm{s}'},
                 'ns': {'prior': {'min': 0.8, 'max': 1.2},
                        'ref': {'dist': 'norm', 'loc': 0.965, 'scale': 0.004},
                        'proposal': 0.002, 'latex': 'n_\\mathrm{s}'}}},
    'SFSR_DESpriors': {
        'desc': 'Adiabatic scalar perturbations, '
                'power law - DESpriors',
        'theory': theory,
        'params': {
            'As_1e9': {'prior': {'min': 0.5, 'max': 5},
                       'ref': {'dist': 'norm', 'loc': 2.1, 'scale': 0.5},
                       'proposal': 0.25, 'latex': '10^9 A_\\mathrm{s})', 'drop': True,
                       'renames': 'A'},
            'As': {'value': 'lambda As_1e9: 1e-9 * As_1e9', 'latex': 'A_\\mathrm{s}'},
            'ns': {'prior': {'min': 0.87, 'max': 1.07},
                   'ref': {'dist': 'norm', 'loc': 0.965, 'scale': 0.05},
                   'proposal': 0.002, 'latex': 'n_\\mathrm{s}'}}},
    'SFSR_lenspriors': {
        'desc': 'Adiabatic scalar perturbations, '
                'power law -- Planck lensing priors',
        'theory': theory,
        'params': {
            'logA': {'prior': {'min': 1.61, 'max': 3.91},
                     'ref': {'dist': 'norm', 'loc': 3.05, 'scale': 0.01},
                     'proposal': 0.005, 'latex': '\\log(10^{10} A_\\mathrm{s})',
                     'drop': True},
            'As': {'value': 'lambda logA: 1e-10*np.exp(logA)', 'latex': 'A_\\mathrm{s}'},
            'ns': {'prior': {'dist': 'norm', 'loc': 0.96, 'scale': 0.02},
                   'ref': {'dist': 'norm', 'loc': 0.965, 'scale': 0.02},
                   'proposal': 0.002, 'latex': 'n_\\mathrm{s}'}}},
    'SFSR_run':
        {'desc': 'Adiabatic scalar perturbations, power law + running spectrum',
         'theory': theory,
         'params': {
             'logA': {'prior': {'min': 1.61, 'max': 3.91},
                      'ref': {'dist': 'norm', 'loc': 3.05, 'scale': 0.001},
                      'proposal': 0.001, 'latex': '\\log(10^{10} A_\\mathrm{s})',
                      'drop': True},
             'As': {'value': 'lambda logA: 1e-10*np.exp(logA)', 'latex': 'A_\\mathrm{s}'},
             'ns': {'prior': {'min': 0.8, 'max': 1.2},
                    'ref': {'dist': 'norm', 'loc': 0.965, 'scale': 0.004},
                    'proposal': 0.002, 'latex': 'n_\\mathrm{s}'},
             'nrun': {'prior': {'min': -1, 'max': 1},
                      'ref': {'dist': 'norm', 'loc': 0, 'scale': 0.005},
                      'proposal': 0.001,
                      'latex': 'n_\\mathrm{run}'}}},
    'SFSR_runrun': {
        'desc': 'Adiabatic scalar perturbations, power law + 2nd-order running spectrum',
        'theory': theory,
        'params': {
            'logA': {'prior': {'min': 1.61, 'max': 3.91},
                     'ref': {'dist': 'norm', 'loc': 3.05, 'scale': 0.001},
                     'proposal': 0.001, 'latex': '\\log(10^{10} A_\\mathrm{s})',
                     'drop': True},
            'As': {'value': 'lambda logA: 1e-10*np.exp(logA)', 'latex': 'A_\\mathrm{s}'},
            'ns': {'prior': {'min': 0.8, 'max': 1.2},
                   'ref': {'dist': 'norm', 'loc': 0.965, 'scale': 0.004},
                   'proposal': 0.002, 'latex': 'n_\\mathrm{s}'},
            'nrun': {'prior': {'min': -1, 'max': 1},
                     'ref': {'dist': 'norm', 'loc': 0, 'scale': 0.005}, 'proposal': 0.001,
                     'latex': 'n_\\mathrm{run}'},
            'nrunrun': {'prior': {'min': -1, 'max': 1},
                        'ref': {'dist': 'norm', 'loc': 0, 'scale': 0.002},
                        'proposal': 0.001, 'latex': 'n_\\mathrm{run,run}'}}},
    'SFSR_t': {
        'desc': 'Adiabatic scalar+tensor perturbations, power law spectrum '
                '(inflation consistency)',
        'theory': {'camb': {'extra_args': {'nt': None}},
                   'classy': {'extra_args': {'n_t': 'scc', 'alpha_t': 'scc'}}},
        'params': {'logA': {'prior': {'min': 1.61, 'max': 3.91},
                            'ref': {'dist': 'norm', 'loc': 3.05, 'scale': 0.001},
                            'proposal': 0.001, 'latex': '\\log(10^{10} A_\\mathrm{s})',
                            'drop': True},
                   'As': {'value': 'lambda logA: 1e-10*np.exp(logA)',
                          'latex': 'A_\\mathrm{s}'},
                   'ns': {'prior': {'min': 0.8, 'max': 1.2},
                          'ref': {'dist': 'norm', 'loc': 0.965, 'scale': 0.004},
                          'proposal': 0.002, 'latex': 'n_\\mathrm{s}'},
                   'r': {'prior': {'min': 0, 'max': 3},
                         'ref': {'dist': 'norm', 'loc': 0, 'scale': 0.03},
                         'proposal': 0.03, 'latex': 'r_{0.05}'}}}, 'SFSR_t_nrun': {
        'desc': 'Adiabatic scalar+tensor perturbations, power law + running '
                'spectrum (inflation consistency)',
        'theory': {'camb': {'extra_args': {'nt': None}},
                   'classy': {'extra_args': {'n_t': 'scc', 'alpha_t': 'scc'}}},
        'params': {'logA': {'prior': {'min': 1.61, 'max': 3.91},
                            'ref': {'dist': 'norm', 'loc': 3.05, 'scale': 0.001},
                            'proposal': 0.001, 'latex': '\\log(10^{10} A_\\mathrm{s})',
                            'drop': True},
                   'As': {'value': 'lambda logA: 1e-10*np.exp(logA)',
                          'latex': 'A_\\mathrm{s}'},
                   'ns': {'prior': {'min': 0.8, 'max': 1.2},
                          'ref': {'dist': 'norm', 'loc': 0.965, 'scale': 0.004},
                          'proposal': 0.002, 'latex': 'n_\\mathrm{s}'},
                   'nrun': {'prior': {'min': -1, 'max': 1},
                            'ref': {'dist': 'norm', 'loc': 0, 'scale': 0.005},
                            'proposal': 0.001, 'latex': 'n_\\mathrm{run}'},
                   'r': {'prior': {'min': 0, 'max': 3},
                         'ref': {'dist': 'norm', 'loc': 0, 'scale': 0.03},
                         'proposal': 0.03, 'latex': 'r_{0.05}'}}}}

# Geometry
geometry = {
    'flat': {'desc': 'Flat FLRW universe',
             'theory': theory},
    'omegak': {'desc': 'FLRW model with varying curvature (prior on Omega_k)',
               'theory': theory,
               'params': {'omegak': {'prior': {'min': -0.3, 'max': 0.3},
                                     'ref': {'dist': 'norm', 'loc': -0.009,
                                             'scale': 0.001},
                                     'proposal': 0.001, 'latex': '\\Omega_k'}}}}

# Hubble parameter constraints
H0_min, H0_max = 20, 100

hubble = {
    'H': {'desc': 'Hubble parameter',
          'theory': theory,
          'params': {'H0': {'prior': {'min': H0_min, 'max': H0_max},
                            'ref': {'dist': 'norm', 'loc': 67, 'scale': 2}, 'proposal': 2,
                            'latex': 'H_0'}}},
    'H_DESpriors': {
        'desc': 'Hubble parameter (reduced range for DES and lensing-only constraints)',
        'theory': theory,
        'params': {
            'H0': {'prior': {'min': 55, 'max': 91},
                   'ref': {'dist': 'norm', 'loc': 67, 'scale': 2}, 'proposal': 2,
                   'latex': 'H_0'}}},
    'sound_horizon_last_scattering': {
        'desc': 'Angular size of the sound horizon at last scattering '
                '(approximate, if using CAMB)',
        'theory': {'camb':
                       {'params':
                            {'theta_MC_100': {'prior': {'min': 0.5, 'max': 10},
                                              'ref': {'dist': 'norm',
                                                      'loc': 1.04109,
                                                      'scale': 0.0004},
                                              'proposal': 0.0002,
                                              'latex': '100\\theta_\\mathrm{MC}',
                                              'drop': True, 'renames': 'theta'},
                             'cosmomc_theta': {
                                 'value': 'lambda theta_MC_100: 1.e-2*theta_MC_100',
                                 'derived': False},
                             'H0': {'latex': 'H_0', 'min': H0_min, 'max': H0_max}},
                        'extra_args': {'theta_H0_range': [H0_min, H0_max]}},
                   'classy': {
                       'params': {
                           'theta_s_1e2': {'prior': {'min': 0.5, 'max': 10},
                                           'ref': {'dist': 'norm', 'loc': 1.0416,
                                                   'scale': 0.0004},
                                           'proposal': 0.0002,
                                           'latex': '100\\theta_\\mathrm{s}',
                                           'drop': True}, '100*theta_s': {
                               'value': 'lambda theta_s_1e2: theta_s_1e2',
                               'derived': False},
                           'H0': {'latex': 'H_0'}}}}},
    'sound_horizon_lensonly': {
        'desc': 'Angular size of the sound horizon (h>0.4; approximate, if using CAMB)',
        'theory': {'camb':
                       {'params':
                            {'theta_MC_100': {'prior': {'min': 0.5, 'max': 10},
                                              'ref': {'dist': 'norm',
                                                      'loc': 1.04109,
                                                      'scale': 0.002},
                                              'proposal': 0.001,
                                              'latex': '100\\theta_\\mathrm{MC}',
                                              'drop': True, 'renames': 'theta'},
                             'cosmomc_theta': {
                                 'value': 'lambda theta_MC_100: 1.e-2*theta_MC_100',
                                 'derived': False},
                             'H0': {'latex': 'H_0', 'min': 40, 'max': H0_max}},
                        'extra_args': {'theta_H0_range': [40, H0_max]}},
                   }},
}

# Matter sector (minus light species)
N_eff_std = 3.044
nu_mass_fac = 94.0708
matter: InfoDict = {
    'omegab_h2, omegac_h2': {
        'desc': 'Flat prior on Omega*h^2 for baryons and cold dark matter',
        'theory': theory,
        'params': {
            'omegabh2': {'prior': {'min': 0.005, 'max': 0.1},
                         'ref': {'dist': 'norm', 'loc': 0.0224, 'scale': 0.0001},
                         'proposal': 0.0001, 'latex': '\\Omega_\\mathrm{b} h^2'},
            'omegach2': {'prior': {'min': 0.001, 'max': 0.99},
                         'ref': {'dist': 'norm', 'loc': 0.12, 'scale': 0.001},
                         'proposal': 0.0005, 'latex': '\\Omega_\\mathrm{c} h^2'},
            'omegam': {'latex': '\\Omega_\\mathrm{m}'}}},
    'Omegab, Omegam':
        {'desc': 'Flat prior on Omega for baryons and total matter',
         'theory': theory,
         'params': {
             'omegab':
                 {'prior': {'min': 0.03, 'max': 0.07},
                  'ref': {'dist': 'norm', 'loc': 0.0495, 'scale': 0.004},
                  'proposal': 0.004, 'latex': '\\Omega_\\mathrm{b}',
                  'drop': True},
             'omegam': {'prior': {'min': 0.1, 'max': 0.9},
                        'ref': {'dist': 'norm',
                                'loc': 0.316,
                                'scale': 0.02},
                        'proposal': 0.02,
                        'latex': '\\Omega_\\mathrm{m}',
                        'drop': True},
             'omegabh2':
                 {'value': 'lambda omegab, H0: omegab*(H0/100)**2',
                  'latex': '\\Omega_\\mathrm{b} h^2'},
             'omegach2': {
                 'value': "lambda omegam, omegab, mnu, H0: (omegam-omegab)*(H0/100)**2"
                          "-(mnu*(%g/3)**0.75)/%g" % (N_eff_std, nu_mass_fac),
                 'latex': '\\Omega_\\mathrm{c} h^2'}}},
    'omegab_h2_lenspriors': {
        'desc': 'BBN-like prior on Omega*h^2 for baryons, with cold dark matter',
        'theory': theory,
        'params': {
            'omegabh2': {'prior': {'dist': 'norm', 'loc': 0.0222, 'scale': 0.0005},
                         'ref': {'dist': 'norm', 'loc': 0.0222, 'scale': 0.0005},
                         'proposal': 0.0004, 'latex': '\\Omega_\\mathrm{b} h^2'},
            'omegach2': {'prior': {'min': 0.001, 'max': 0.99},
                         'ref': {'dist': 'norm', 'loc': 0.12, 'scale': 0.003},
                         'proposal': 0.002, 'latex': '\\Omega_\\mathrm{c} h^2'},
            'omegam': {'latex': '\\Omega_\\mathrm{m}'}}},
}

for m in matter.values():
    m["params"]["omegamh2"] = {"derived": "lambda omegam, H0: omegam*(H0/100)**2",
                               "latex": r"\Omega_\mathrm{m} h^2"}

# Neutrinos and other extra matter
neutrinos: InfoDict = {
    'one_heavy_planck':
        {'desc': 'Two massless nu and one with m=0.06. Neff=3.044',
         'theory': {
             'camb': {
                 'extra_args': {'num_massive_neutrinos': 1, 'nnu': 3.044},
                 'params': {'mnu': 0.06}},
             'classy': {
                 'extra_args': {'N_ncdm': 1, 'N_ur': 2.0328},
                 'params': {'m_ncdm': {'value': 0.06, 'renames': 'mnu'}}}}},
    'varying_mnu':
        {'desc': "Varying total mass of 3 degenerate nu's, with N_eff=3.044",
         'theory': {
             'camb': {
                 'extra_args': {'num_massive_neutrinos': 3, 'nnu': 3.044},
                 'params': {'mnu': {'prior': {'min': 0, 'max': 5},
                                    'ref': {'dist': 'norm', 'loc': 0.02, 'scale': 0.1},
                                    'proposal': 0.03,
                                    'latex': '\\sum m_\\nu'}}},
             'classy': {
                 'extra_args': {'N_ncdm': 1, 'deg_ncdm': 3, 'N_ur': 0.00641},
                 'params':
                     {'m_ncdm': {'prior': {'min': 0, 'max': 1.667},
                                 'ref': {'dist': 'norm', 'loc': 0.0067, 'scale': 0.033},
                                 'proposal': 0.01, 'latex': 'm_\\nu'},
                      'mnu': {'derived': 'lambda m_ncdm: 3 * m_ncdm',
                              'latex': '\\sum m_\\nu'}}}}},
    'varying_Neff':
        {'desc': 'Varying Neff with two massless nu and one with m=0.06',
         'theory': {
             'camb': {
                 'extra_args': {'num_massive_neutrinos': 1},
                 'params': {
                     'mnu': 0.06,
                     'nnu': {
                         'prior': {'min': 0.05, 'max': 10},
                         'ref': {'dist': 'norm', 'loc': 3.044, 'scale': 0.05},
                         'proposal': 0.05,
                         'latex': 'N_\\mathrm{eff}'}}},
             'classy':
                 {'extra_args': {'N_ncdm': 1},
                  'params': {
                      'm_ncdm': {'value': 0.06, 'renames': 'mnu'},
                      'N_ur': {'prior': {'min': 0.0001, 'max': 9},
                               'ref': {'dist': 'norm', 'loc': 2.0328,
                                       'scale': 0.05}, 'proposal': 0.05,
                               'latex': 'N_\\mathrm{ur}'},
                      'nnu': {'derived': 'lambda Neff: Neff',
                              'latex': 'N_\\mathrm{eff}'}}}}},
    'varying_mnu_Neff':
        {'desc': "Varying Neff and total mass of 3 degenerate nu's",
         'theory':
             {'camb': {
                 'extra_args': {'num_massive_neutrinos': 3},
                 'params': {
                     'mnu': {'prior': {'min': 0, 'max': 5},
                             'ref': {'dist': 'norm', 'loc': 0.02, 'scale': 0.1},
                             'proposal': 0.03,
                             'latex': '\\sum m_\\nu'},
                     'nnu': {
                         'prior': {'min': 0.05, 'max': 10},
                         'ref': {'dist': 'norm', 'loc': 3.044, 'scale': 0.05},
                         'proposal': 0.05,
                         'latex': 'N_\\mathrm{eff}'}}}}}}
# Dark Energy
dark_energy: InfoDict = \
    {'lambda':
         {'desc': 'Cosmological constant (w=-1)',
          'theory': theory,
          'params': {'omegal': {'latex': '\\Omega_\\Lambda'}}},
     'de_w':
         {'desc': 'Varying constant eq of state',
          'theory': {'camb': None, 'classy': {'params': {'Omega_Lambda': 0}}},
          'params': {
              'w': {'prior': {'min': -3, 'max': -0.333},
                    'ref': {'dist': 'norm', 'loc': -0.99, 'scale': 0.02},
                    'proposal': 0.02,
                    'latex': 'w_\\mathrm{DE}'}}},
     'de_w_wa':
         {'desc': 'Varying constant eq of state with w(a) = w0 + (1-a) wa',
          'theory': {'camb': {'extra_args': {'dark_energy_model': 'ppf'}},
                     'classy': {'params': {'Omega_Lambda': 0}}},
          'params': {
              'w': {'prior': {'min': -3, 'max': 1},
                    'ref': {'dist': 'norm', 'loc': -0.99, 'scale': 0.02},
                    'proposal': 0.02,
                    'latex': 'w_{0,\\mathrm{DE}}'},
              'wa': {'prior': {'min': -3, 'max': 2},
                     'ref': {'dist': 'norm', 'loc': 0, 'scale': 0.05},
                     'proposal': 0.05,
                     'latex': 'w_{a,\\mathrm{DE}}'}}}}

# BBN
bbn_derived_camb: InfoDict = {'YpBBN': {'latex': 'Y_P^\\mathrm{BBN}'},
                              'DHBBN': {'derived': 'lambda DH: 10**5*DH',
                                        'latex': '10^5 \\mathrm{D}/\\mathrm{H}'}}
bbn = {'consistency':
           {'desc': 'Primordial Helium fraction inferred from BBN consistency',
            'theory': {
                'camb': {'params': bbn_derived_camb},
                'classy': None},
            'params': {'yheused': {'latex': 'Y_\\mathrm{P}'}}},
       'YHe_des_y1':
           {'desc': 'Fixed Y_P = 0.245341 (used in DES Y1)',
            'theory': theory,
            'params': {'yhe': 0.245341}},
       'YHe':
           {'desc': 'Varying primordial Helium fraction',
            'theory': theory,
            'params': {
                'yhe': {'prior': {'min': 0.1, 'max': 0.5},
                        'ref': {'dist': 'norm', 'loc': 0.237, 'scale': 0.006},
                        'proposal': 0.006, 'latex': 'Y_\\mathrm{P}'}}}}

# Reionization
reionization = {
    'std': {
        'desc': 'Standard reio, lasting delta_z=0.5',
        'theory': theory,
        'params': {
            'tau': {'prior': {'min': 0.01, 'max': 0.8},
                    'ref': {'dist': 'norm', 'loc': 0.055, 'scale': 0.006},
                    'proposal': 0.003,
                    'latex': '\\tau_\\mathrm{reio}'},
            'zrei': {'latex': 'z_\\mathrm{re}'}}},
    'gauss_prior': {
        'desc': 'Standard reio, lasting delta_z=0.5, gaussian prior around tau=0.07',
        'theory': theory,
        'params': {
            'tau': {'prior': {'dist': 'norm', 'loc': 0.07, 'scale': 0.02},
                    'ref': {'dist': 'norm', 'loc': 0.07, 'scale': 0.01},
                    'proposal': 0.005,
                    'latex': '\\tau_\\mathrm{reio}'},
            'zrei': {'latex': 'z_\\mathrm{re}'}}},
    'irrelevant': {
        'desc': 'Irrelevant (NB: only valid for non-CMB or CMB-marged datasets!)',
        'theory': theory,
        'params': {}}}

# EXPERIMENTS ############################################################################
base_precision: InfoDict = {"camb": None,
                            "classy": {"non linear": "hmcode", "nonlinear_min_k_max": 20}}
cmb_precision = deepcopy(base_precision)
cmb_precision["camb"] = {"lens_potential_accuracy": 1}

planck_lss_precision = deepcopy(base_precision)
planck_lss_precision["camb"] = {"halofit_version": "mead",
                                "bbn_predictor": "PArthENoPE_880.2_standard.dat"}

planck_precision = deepcopy(planck_lss_precision)
planck_precision["camb"]["lens_potential_accuracy"] = 1

default_mcmc_options = {"proposal_scale": 1.9,
                        "Rminus1_stop": 0.01, "Rminus1_cl_stop": 0.2}
cmb_sampler_recommended: InfoDict = {"mcmc": dict(drag=True, oversample_power=0.4,
                                                  **default_mcmc_options)}
cmb_sampler_mcmc: InfoDict = {"mcmc": dict(drag=False, **default_mcmc_options)}

like_cmb: InfoDict = {
    none: {},
    "planck_NPIPE": {
        "desc": "Planck NPIPE (native; polarized NPIPE CMB + lensing)",
        "sampler": cmb_sampler_recommended,
        "theory": {theo: {"extra_args": cmb_precision[theo]}
                   for theo in ["camb", "classy"]},
        "likelihood": {
            "planck_2018_lowl.TT": None,
            "planck_2018_lowl.EE": None,
            "planck_NPIPE_highl_CamSpec.TTTEEE": None,
            "planckpr4lensing":
                {'package_install': {'github_repository': 'carronj/planck_PR4_lensing',
                                     'min_version': '1.0.2'}}}},
    "planck_2018": {
        "desc": "Planck 2018 (Polarized CMB + lensing)",
        "sampler": cmb_sampler_recommended,
        "theory": {theo: {"extra_args": planck_precision[theo]}
                   for theo in ["camb", "classy"]},
        "likelihood": {
            "planck_2018_lowl.TT": None,
            "planck_2018_lowl.EE": None,
            "planck_2018_highl_plik.TTTEEE": None,
            "planck_2018_lensing.clik": None}},
    "planck_2018_bk18": {
        "desc": "Planck 2018 (Polarized CMB + lensing) + Bicep/Keck-Array 2018",
        "sampler": cmb_sampler_recommended,
        "theory": {theo: {"extra_args": planck_precision[theo]}
                   for theo in ["camb", "classy"]},
        "likelihood": {
            "planck_2018_lowl.TT": None,
            "planck_2018_lowl.EE": None,
            "planck_2018_highl_plik.TTTEEE": None,
            "planck_2018_lensing.clik": None,
            "bicep_keck_2018": None}},
    "planck_2018_CMBmarged_lensing": {
        "desc": "Planck 2018 CMB-marginalized lensing only",
        "sampler": cmb_sampler_mcmc,
        "theory": {theo: {"extra_args": planck_precision[theo]}
                   for theo in ["camb", "classy"]},
        "likelihood": {"planck_2018_lensing.CMBMarged": None}}}

# Add common CMB derived parameters
derived_params = {'sigma8': {'latex': '\\sigma_8'},
                  's8h5': {'derived': 'lambda sigma8, H0: sigma8*(H0*1e-2)**(-0.5)',
                           'latex': '\\sigma_8/h^{0.5}'},
                  's8omegamp5': {'derived': 'lambda sigma8, omegam: sigma8*omegam**0.5',
                                 'latex': '\\sigma_8 \\Omega_\\mathrm{m}^{0.5}'},
                  's8omegamp25': {'derived': 'lambda sigma8, omegam: sigma8*omegam**0.25',
                                  'latex': '\\sigma_8 \\Omega_\\mathrm{m}^{0.25}'},
                  'A': {'derived': 'lambda As: 1e9*As', 'latex': '10^9 A_\\mathrm{s}'},
                  'clamp': {'derived': 'lambda As, tau: 1e9*As*np.exp(-2*tau)',
                            'latex': '10^9 A_\\mathrm{s} e^{-2\\tau}'},
                  'age': {'latex': '{\\rm{Age}}/\\mathrm{Gyr}'},
                  'rdrag': {'latex': 'r_\\mathrm{drag}'}
                  }
for name, m in like_cmb.items():
    # Don't add the derived parameter to the no-CMB case!
    if not m:
        continue
    if "params" not in m:
        m["params"] = dict()
    m["params"].update(derived_params)
    if "cmbmarged" in name.lower():
        m["params"].pop("A")
        m["params"].pop("clamp")
# Some more, in case we want to add them at some point, described in
# https://wiki.cosmos.esa.int/planckpla2015/images/b/b9/Parameter_tag_definitions_2015.pdf
#    "zstar":       {"latex": r"z_*"},
#    "rstar":       {"latex": r"r_*"},
#    "DAstar":      {"latex": r"D_\mathrm{A}/\mathrm{Gpc}"},
#    "zdrag":       {"latex": r"z_\mathrm{drag}"},
#    "kd":          {"latex": r"k_\mathrm{D}"},
#    "thetad":      {"latex": r"100\theta_\mathrm{D}"},
#    "zeq":         {"latex": r"z_\mathrm{eq}"},
#    "keq":         {"latex": r"k_\mathrm{eq}"},
#    "thetaeq":     {"latex": r"100\theta_\mathrm{eq}"},
#    "thetastar":   {"latex": r"100\theta_*"},
#    "thetarseq":   {"latex": r"100\theta_\mathrm{s,eq}"},

like_bao = {none: {},
            'BAO_planck_2018': {
                'desc': 'Baryon acoustic oscillation data from DR12, MGS and 6DF '
                        '(Planck 2018 papers)',
                'theory': theory,
                'likelihood': {'bao.sixdf_2011_bao': None, 'bao.sdss_dr7_mgs': None,
                               'bao.sdss_dr12_consensus_bao': None}},
            'BAO_planck_latest': {
                'desc': 'Baryon acoustic oscillation data from BOSS DR12, eBOSS DR16, '
                        'MGS and 6DF',
                'theory': theory,
                'likelihood': {'bao.sixdf_2011_bao': None, 'bao.sdss_dr7_mgs': None,
                               'bao.sdss_dr16_baoplus_lrg': None,
                               'bao.sdss_dr16_baoplus_elg': None,
                               'bao.sdss_dr16_baoplus_qso': None,
                               'bao.sdss_dr16_baoplus_lyauto': None,
                               'bao.sdss_dr16_baoplus_lyxqso': None,
                               }}}

like_des: InfoDict = \
    {none: {},
     "des_y1_clustering": {
         "desc": "Galaxy clustering from DES Y1",
         "likelihood": {"des_y1.clustering": None}},
     "des_y1_galaxy_galaxy": {
         "desc": "Galaxy-galaxy lensing from DES Y1",
         "likelihood": {"des_y1.galaxy_galaxy": None}},
     "des_y1_shear": {
         "desc": "Cosmic shear data from DES Y1",
         "likelihood": {"des_y1.shear": None}},
     "des_y1_joint": {
         "desc": "Combination of galaxy clustering and weak lensing data from DES Y1",
         "likelihood": {"des_y1.joint": None}}}

for key, value in like_des.items():
    if key is not none:
        value['theory'] = {theo: {"extra_args": base_precision[theo]}
                           for theo in ["camb", "classy"]}
        value['sampler'] = cmb_sampler_recommended

like_sn: InfoDict = {none: {},
                     "Pantheon": {
                         "desc": "Supernovae data from the Pantheon sample",
                         "theory": theory,
                         "likelihood": {"sn.pantheon": None}}}

like_H0: InfoDict = \
    {none: {},
     "Riess2018a": {
         "desc": "Local H0 measurement from Riess et al. 2018a (used in Planck 2018)",
         "theory": theory,
         "likelihood": {"H0.riess2018a": None}},
     "Riess201903": {
         "desc": "Local H0 measurement from Riess et al. 2019",
         "theory": theory,
         "likelihood": {"H0.riess201903": None}},
     "Riess2020": {
         "desc": "Local H0 measurement from Riess et al. 2020",
         "theory": theory,
         "likelihood": {"H0.riess2020": None}},
     "Freedman2020": {
         "desc": "Local H0 measurement from Freedman et al. 2020",
         "theory": theory,
         "likelihood": {"H0.freedman2020": None}},
     "Riess2020Mb": {
         "desc": "Local magnitude measurement as from Riess et al. 2020",
         "theory": theory,
         "likelihood": {"H0.riess2020Mb": None, "sn.pantheon": {"use_abs_mag": True}}},
     }

# SAMPLERS ###############################################################################

sampler: InfoDict = {
    "MCMC":
        {"desc": "MCMC sampler with covmat learning",
         "sampler": {"mcmc": {"covmat": "auto"}}},
    "MCMC dragging":
        {"desc": "MCMC fast-dragging sampler with covmat learning",
         "sampler": {"mcmc": {"drag": True, "oversample_power": 0.4,
                              "proposal_scale": 1.9, "covmat": "auto"}}},
    "PolyChord": {
        "desc": "Nested sampler, affine invariant and multi-modal",
        "sampler": {"polychord": None}}}

# PRESETS ################################################################################

planck_base_model = {
    "primordial": "SFSR",
    "geometry": "flat",
    "hubble": "sound_horizon_last_scattering",
    "matter": "omegab_h2, omegac_h2",
    "neutrinos": "one_heavy_planck",
    "dark_energy": "lambda",
    "bbn": "consistency",
    "reionization": "std"}
default_sampler = {"sampler": "MCMC dragging"}

preset: InfoDict = dict([
    (none, {"desc": "(No preset chosen)"}),
    # Pure CMB #######################################################
    ("planck_NPIPE_camb", {
        "desc": "Planck NPIPE with CAMB (all native Python)",
        "theory": "camb",
        "like_cmb": "planck_NPIPE"}),
    ("planck_2018_camb", {
        "desc": "Planck 2018 with CAMB",
        "theory": "camb",
        "like_cmb": "planck_2018"}),
    ("planck_2018_classy", {
        "desc": "Planck 2018 with CLASS",
        "theory": "classy",
        "like_cmb": "planck_2018"}),
    ("planck_2018_bicep_camb", {
        "desc": "Planck 2018 + BK18 (with tensor modes) with CAMB",
        "theory": "camb",
        "primordial": "SFSR_t",
        "like_cmb": "planck_2018_bk18"}),
    ("planck_2018_bicep_classy", {
        "desc": "Planck 2018 + BK18 (with tensor modes) with CLASS",
        "theory": "classy",
        "primordial": "SFSR_t",
        "like_cmb": "planck_2018_bk18"}),
    # CMB+BAO ######################################################
    ("planck_2018_BAO_camb", {
        "desc": "Planck 2018 + BAO with CAMB",
        "theory": "camb",
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_2018"}),
    ("planck_2018_BAO_classy", {
        "desc": "Planck 2018 + BAO with CLASS",
        "theory": "classy",
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_2018"}),
    ("planck_BAO_latest_camb", {
        "desc": "Planck 2018 + eBOSS 16 BAO with CAMB",
        "theory": "camb",
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_latest"}),
    ("planck_BAO_latest_classy", {
        "desc": "Planck 2018 + eBOSS 16 BAO with CLASS",
        "theory": "classy",
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_latest"}),

    # CMB+BAO+SN ###################################################
    ("planck_2018_BAO_SN_camb", {
        "desc": "Planck 2018 + BAO + SN with CAMB",
        "theory": "camb",
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_latest",
        "like_sn": "Pantheon"}),
    ("planck_2018_BAO_SN_classy", {
        "desc": "Planck 2018 + BAO + SN with CLASS",
        "theory": "classy",
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_latest",
        "like_sn": "Pantheon"}),
    # CMB+DES+BAO+SN ###################################################
    ("planck_2018_DES_BAO_SN_camb", {
        "desc": "Planck 2018 + DESjoint + BAO + SN with CAMB",
        "theory": "camb",
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_latest",
        "like_des": "des_y1_joint",
        "like_sn": "Pantheon"}),
    ("planck_2018_DES_BAO_SN_classy", {
        "desc": "Planck 2018 + DESjoint + BAO + SN with CLASS",
        "theory": "classy",
        "like_cmb": "planck_2018",
        "like_bao": "BAO_planck_latest",
        "like_des": "des_y1_joint",
        "like_sn": "Pantheon"}),
])

# Add planck baseline model
for pre in preset.values():
    pre.update(
        {field: value for field, value in planck_base_model.items() if field not in pre})

# Lensing-only ###################################################
preset.update({
    none: {"desc": "(No preset chosen)"},
    "planck_2018_lensonly_camb": {
        "desc": "Planck 2018 lensing only with CAMB",
        "theory": "camb",
        "like_cmb": "planck_2018_CMBmarged_lensing",
        "like_des": none,
        "primordial": "SFSR_lenspriors",
        "geometry": "flat",
        "hubble": "sound_horizon_lensonly",
        "matter": "omegab_h2_lenspriors",
        "neutrinos": "one_heavy_planck",
        "dark_energy": "lambda",
        "bbn": "consistency",
        "reionization": "irrelevant",
        "sampler": "MCMC"
    },
    "planck_2018_DES_lensingonly_camb": {
        "desc": "Planck 2018 + DES Y1 lensing-only with CAMB",
        "theory": "camb",
        "like_cmb": "planck_2018_CMBmarged_lensing",
        "like_des": "des_y1_shear"},
    "planck_2018_DES_lensingonly_classy": {
        "desc": "Planck 2018 + DES Y1 lensing-only with CLASS",
        "theory": "classy",
        "like_cmb": "planck_2018_CMBmarged_lensing",
        "like_des": "des_y1_shear"}})

lensingonly_DES_model = {
    "primordial": "SFSR_DESpriors",
    "geometry": "flat",
    "hubble": "H_DESpriors",
    "matter": "Omegab, Omegam",
    "neutrinos": "one_heavy_planck",
    "dark_energy": "lambda",
    "bbn": "YHe_des_y1",
    "reionization": "irrelevant"
}

# Add planck baseline model
for name, pre in preset.items():
    if "lensingonly" in name:
        pre.update(
            {field: value for field, value in lensingonly_DES_model.items() if
             field not in pre})
    if pre.get("sampler") != 'MCMC':
        pre.update(default_sampler)

# BASIC INSTALLATION #####################################################################
install_basic: InfoDict = {
    "theory": theory,
    "likelihood": {
        # Native first: avoids reinstalling clik code+data if supp data obsolete
        "planck_2018_lensing.native": None,
        "planck_2018_lowl.TT": None,
        "planck_2018_lowl.EE": None,
        "sn.pantheon": None,
        "bao.sdss_dr12_consensus_final": None,
        "des_y1.joint": None}}

install_tests = deepcopy(install_basic)
install_tests["likelihood"].update({"planck_2015_lowl": None,
                                    "planck_2018_highl_plik.TT_unbinned": None,
                                    "planck_2018_highl_plik.TT_lite_native": None,
                                    "planck_2018_highl_CamSpec.TT": None,
                                    "planck_2018_highl_CamSpec2021.TT": None,
                                    "bicep_keck_2018": None
                                    })

# CONTENTS FOR COMBO-BOXED IN A GUI ######################################################

_combo_dict_text = (
    ["Presets", (["preset", "Presets"],)],
    ["Cosmological Model", (
        ["theory", "Theory code"],
        ["primordial", "Primordial perturbations"],
        ["geometry", "Geometry"],
        ["hubble", "Hubble parameter constraint"],
        ["matter", "Matter sector"],
        ["neutrinos", "Neutrinos and other extra matter"],
        ["dark_energy", "Lambda / Dark energy"],
        ["bbn", "BBN"],
        ["reionization", "Reionization history"])],
    ["Data sets", (
        ["like_cmb", "CMB experiments"],
        ["like_bao", "BAO experiments"],
        ["like_des", "DES measurements"],
        ["like_sn", "SN experiments"],
        ["like_H0", "Local H0 measurements"])],
    ["Sampler", (["sampler", "Samplers"],)])



"""
.. module:: run

:Synopsis: Main scripts to run the sampling
:Author: Jesus Torrado

"""

# Global
from typing import Union, Optional, Tuple
import os

# Local
from cobaya.conventions import packages_path_arg, packages_path_arg_posix, get_version, \
    packages_path_input
from cobaya.typing import InputDict, LiteralFalse
from cobaya.output import get_output
from cobaya.model import Model
from cobaya.sampler import get_sampler_name_and_class, check_sampler_info, Sampler
from cobaya.log import logger_setup, is_debug, get_logger, LoggedError
from cobaya.yaml import yaml_dump
from cobaya.input import update_info, load_info_overrides
from cobaya.tools import warn_deprecation, recursive_update, sort_cosmetic
from cobaya.post import post, PostResultDict
from cobaya import mpi


def run(info_or_yaml_or_file: Union[InputDict, str, os.PathLike],
        packages_path: Optional[str] = None,
        output: Union[str, LiteralFalse, None] = None,
        debug: Union[bool, int, None] = None,
        stop_at_error: Optional[bool] = None,
        resume: bool = None, force: bool = None,
        minimize: Optional[bool] = None,
        no_mpi: bool = False, test: bool = None,
        override: Optional[InputDict] = None,
        ) -> Tuple[InputDict, Union[Sampler, PostResultDict]]:
    """
    Run from an input dictionary, file name or yaml string, with optional arguments
    to override settings in the input as needed.

    :param info_or_yaml_or_file: input options dictionary, yaml file, or yaml text
    :param packages_path: path where external packages were installed
    :param output: path name prefix for output files, or False for no file output
    :param debug: true for verbose debug output, or a specific logging level
    :param stop_at_error: stop if an error is raised
    :param resume: continue an existing run
    :param force: overwrite existing output if it exists
    :param minimize: if true, ignores the sampler and runs default minimizer
    :param no_mpi: run without MPI
    :param test: only test initialization rather than actually running
    :param override: option dictionary to merge into the input one, overriding settings
       (but with lower precedence than the explicit keyword arguments)
    :return: (updated_info, sampler) tuple of options dictionary and Sampler instance,
              or (updated_info, post_results) if using "post" post-processing
    """
    # This function reproduces the model-->output-->sampler pipeline one would follow
    # when instantiating by hand, but alters the order to perform checks and dump info
    # as early as possible, e.g. to check if resuming possible or `force` needed.
    if no_mpi or test:
        mpi.set_mpi_disabled()
    with mpi.ProcessState("run"):
        flags = {packages_path_input: packages_path, "debug": debug,
                 "stop_at_error": stop_at_error, "resume": resume, "force": force,
                 "minimize": minimize, "test": test}
        info: InputDict = load_info_overrides(
            info_or_yaml_or_file, override or {}, **flags)
        if info.get("post"):
            if info.get("minimize"):  # type: ignore
                raise ValueError(
                    "``minimize`` option is incompatible with post-processing.")
            if isinstance(output, str) or output is False:
                info["post"]["output"] = output or None
            return post(info)
        # Set up output and logging
        if isinstance(output, str) or output is False:
            info["output"] = output or None
        # MARKED FOR DEPRECATION IN v3.2
        if info.get("debug_file"):
            raise LoggedError("'debug_file' has been deprecated. If you want to "
                              "save the debug output to a file, use 'debug: [filename]'.")
        # END OF DEPRECATION BLOCK
        logger_setup(info.get("debug"))
        logger_run = get_logger(run.__name__)
        # 1. Prepare output driver, if requested by defining an output prefix
        # GetDist needs to know the original sampler, so don't overwrite if minimizer
        try:
            which_sampler = list(info["sampler"])[0]
            if info.get("minimize"):  # type: ignore
                # Preserve options if "minimize" was already the sampler
                if which_sampler.lower() != "minimize":
                    info["sampler"] = {"minimize": None}
                    which_sampler = "minimize"
        except (KeyError, TypeError):
            raise LoggedError(
                logger_run, "You need to specify a sampler using the 'sampler' key "
                            "as e.g. `sampler: {mcmc: None}.`")
        infix = "minimize" if which_sampler == "minimize" else None
        with get_output(prefix=info.get("output"), resume=info.get("resume"),
                        force=info.get("force"), infix=infix) as out:
            # 2. Update the input info with the defaults for each component
            updated_info = update_info(info)
            if is_debug(logger_run):
                # Dump only if not doing output
                # (otherwise, the user can check the .updated file)
                if not out and mpi.is_main_process():
                    logger_run.info(
                        "Input info updated with defaults (dumped to YAML):\n%s",
                        yaml_dump(sort_cosmetic(updated_info)))
            # 3. If output requested, check compatibility if existing one, and dump.
            # 3.1 First: model only
            out.check_and_dump_info(info, updated_info, cache_old=True,
                                    ignore_blocks=["sampler"])
            # 3.2 Then sampler -- 1st get the last sampler mentioned in the updated.yaml
            # TODO: ideally, using Minimizer would *append* to the sampler block.
            #       Some code already in place, but not possible at the moment.
            try:
                last_sampler = list(updated_info["sampler"])[-1]
                last_sampler_info = {last_sampler: updated_info["sampler"][last_sampler]}
            except (KeyError, TypeError):
                raise LoggedError(logger_run, "No sampler requested.")
            sampler_name, sampler_class = get_sampler_name_and_class(last_sampler_info)
            check_sampler_info(
                (out.reload_updated_info(use_cache=True) or {}).get("sampler"),
                updated_info["sampler"], is_resuming=out.is_resuming())
            # Dump again, now including sampler info
            out.check_and_dump_info(info, updated_info, check_compatible=False)
            # Check if resumable run
            sampler_class.check_force_resume(
                out, info=updated_info["sampler"][sampler_name])
            # 4. Initialize the posterior and the sampler
            with Model(updated_info["params"], updated_info["likelihood"],
                       updated_info.get("prior"), updated_info.get("theory"),
                       packages_path=info.get(packages_path_input),
                       timing=updated_info.get("timing"),
                       allow_renames=False,
                       stop_at_error=info.get("stop_at_error", False)) as model:
                # Re-dump the updated info, now containing parameter routes and version
                updated_info = recursive_update(updated_info, model.info())
                out.check_and_dump_info(None, updated_info, check_compatible=False)
                sampler = sampler_class(updated_info["sampler"][sampler_name],
                                        model, out, name=sampler_name,
                                        packages_path=info.get(packages_path_input))
                # Re-dump updated info, now also containing updates from the sampler
                updated_info["sampler"][sampler_name] = \
                    recursive_update(updated_info["sampler"][sampler_name],
                                     sampler.info())
                out.check_and_dump_info(None, updated_info, check_compatible=False)
                mpi.sync_processes()
                if info.get("test", False):
                    logger_run.info("Test initialization successful! "
                                    "You can probably run now without `--%s`.", "test")
                    return updated_info, sampler
                # Run the sampler
                sampler.run()
    return updated_info, sampler


# Command-line script
def run_script(args=None):
    """Shell script wrapper for :func:`run.run` (including :func:`post.post`)"""
    warn_deprecation()
    import argparse
    # kwargs for flags that should be True|None, instead of True|False
    # (needed in order not to mistakenly override input file inside the run() function
    trueNone_kwargs = {"action": "store_true", "default": None}
    parser = argparse.ArgumentParser(
        prog="cobaya run", description="Cobaya's run script.")
    parser.add_argument("input_file", action="store", metavar="input_file.yaml",
                        help="An input file to run.")
    parser.add_argument("-" + packages_path_arg[0], "--" + packages_path_arg_posix,
                        action="store", metavar="/packages/path", default=None,
                        help="Path where external packages were installed.")
    parser.add_argument("-" + "o", "--" + "output",
                        action="store", metavar="/some/path", default=None,
                        help="Path and prefix for the text output.")
    parser.add_argument("-" + "d", "--" + "debug",
                        help="Produce verbose debug output.", **trueNone_kwargs)
    continuation = parser.add_mutually_exclusive_group(required=False)
    continuation.add_argument("-" + "r", "--" + "resume",
                              help=("Resume an existing chain if it has similar info "
                                    "(fails otherwise)."), **trueNone_kwargs)
    continuation.add_argument("-" + "f", "--" + "force",
                              help=("Overwrites previous output, if it exists "
                                    "(use with care!)"), **trueNone_kwargs)
    parser.add_argument("--%s" % "test",
                        help="Initialize model and sampler, and exit.", **trueNone_kwargs)
    parser.add_argument("-M", "--minimize",
                        help=("Replaces the sampler in the input and runs a minimization "
                              "process (incompatible with post-processing)."),
                        **trueNone_kwargs)
    parser.add_argument("--version", action="version", version=get_version())
    parser.add_argument("--no-mpi",
                        help=("disable MPI when mpi4py installed but MPI does "
                              "not actually work"),
                        **trueNone_kwargs)
    arguments = parser.parse_args(args)
    info = arguments.input_file
    del arguments.input_file
    if not info.endswith('.yaml') and not os.path.exists(info):
        if os.path.exists(info + '.yaml'):
            info = info + '.yaml'
    run(info, **arguments.__dict__)


if __name__ == '__main__':
    run_script()



from cobaya.likelihoods.base_classes import PlanckClik


class planck_2015_lensing(PlanckClik):
    r"""
    Lensing likelihood of Planck's 2015 data release based on temperature+polarization
    map-based lensing reconstruction \cite{Ade:2015zua}.
    """
    pass



r"""
.. module:: Mb

:Synopsis: Prototype class for local Hubble parameter measurements quantified in terms
    of the magnitude measurement (closer to what is measured than H0)
:Author: Pablo Lemos
"""

from cobaya.likelihood import Likelihood


class Mb(Likelihood):
    # Data type for aggregated chi2 (case sensitive)
    type = "Mb"

    # variables from yaml
    Mb_mean: float
    Mb_std: float

    def initialize(self):
        self.minus_half_invvar = - 0.5 / self.Mb_std ** 2

    def get_requirements(self):
        return {}

    def logp(self, **params_values):
        Mb_theory = params_values.get("Mb", None)
        return self.minus_half_invvar * (Mb_theory - self.Mb_mean) ** 2



"""
Automatic tests of the quickstart example in the documentation,
to make sure it remains up to date.
"""

from flaky import flaky
from io import StringIO
import os
import pytest

from cobaya.yaml import yaml_load_file
from cobaya.input import is_equal_info
from cobaya.tools import KL_norm
from .common_sampler import KL_tolerance
from .common import stdout_redirector
from cobaya import mpi

pytestmark = pytest.mark.mpi

docs_folder = os.path.join(os.path.dirname(os.path.realpath(__file__)), "..", "docs")
docs_src_folder = os.path.join(docs_folder, "src_examples", "quickstart")
docs_img_folder = os.path.join(docs_folder, "img")


@flaky(max_runs=3, min_passes=1)
@mpi.sync_errors
def test_example():
    # temporarily change working directory to be able to run the files "as is"
    cwd = os.getcwd()
    try:
        os.chdir(docs_src_folder)
        info_yaml = yaml_load_file("gaussian.yaml")
        info_yaml.pop("output")
        globals_example = {}
        exec(open(os.path.join(docs_src_folder, "create_info.py")).read(),
             globals_example)
        assert is_equal_info(info_yaml, globals_example["info"]), (
            "Inconsistent info between yaml and interactive.")
        exec(open(os.path.join(docs_src_folder, "load_info.py")).read(), globals_example)
        globals_example["info_from_yaml"].pop("output")
        assert is_equal_info(info_yaml, globals_example["info_from_yaml"]), (
            "Inconsistent info between interactive and *loaded* yaml.")
        # Run the chain -- constant seed so results are the same!
        globals_example["info"]["sampler"]["mcmc"] = (
                globals_example["info"]["sampler"]["mcmc"] or {})
        exec(open(os.path.join(docs_src_folder, "run.py")).read(), globals_example)
        # Run the minimizer -- output doesn't matter. Just checking that it does not fail
        exec(open(os.path.join(docs_src_folder, "run_min.py")).read(), globals_example)
        # Analyze and plot -- capture print output
        stream = StringIO()
        with stdout_redirector(stream):
            exec(open(os.path.join(docs_src_folder, "analyze.py")).read(),
                 globals_example)
        # Checking results
        mean, covmat = [globals_example["info"]["likelihood"]["gaussian_mixture"][x]
                        for x in ["means", "covs"]]
        assert (KL_norm(
            m1=mean, S1=covmat, m2=globals_example["mean"], S2=globals_example["covmat"])
                <= KL_tolerance), (
            "Sampling appears not to have worked too well. Run again?")
    finally:
        # Back to the working directory of the tests, just in case
        os.chdir(cwd)



from cobaya.likelihoods.base_classes import PlanckPlikLite


class TTTEEE_lite_native(PlanckPlikLite):
    r"""
    High-$\ell$ temperature and polarization, foreground-marginalized likelihood of
    Planck's 2018 data release \cite{Aghanim:2019ame} (Python re-implementation of the
    official \textsc{plik\_lite}, by E.~Calabrese, and A.~Lewis).
    """
    pass



"""
Tests for the PoolXD classes, which are used to assist caching of computed quantities
which are a function of a finite set of fixed values.
"""

import pytest
import numpy as np
from flaky import flaky

from cobaya.tools import Pool1D, Pool2D

# Max number of tries per test
max_runs = 3

# Number of pool and test points
n_pool = 500
n_test = 100
# size of the perturbation rel factor and test abs factor
r_perturb = 1e-16
a_tol_test = 1e-8


@flaky(max_runs=max_runs, min_passes=1)
def test_pool1d():
    values = np.random.random(n_pool)
    pool = Pool1D(values)
    test_values = np.random.choice(values, n_test) + r_perturb * np.random.random(n_test)
    # At least a duplicate, for robustness
    test_values[-1] = test_values[0]
    indices = pool.find_indices(test_values)
    assert(np.all(np.abs(test_values - pool[indices] < a_tol_test)))


def test_pool1d_fail():
    values = np.random.random(1)
    pool = Pool1D(values)
    test_values = [2]  # out of range
    with pytest.raises(ValueError):
        pool.find_indices(test_values)


@flaky(max_runs=max_runs, min_passes=1)
def test_pool2d(from_list=False):
    if from_list:
        # num of combinations N needs to be ~= n_test
        # if list of length m (large): N = (m**2 - m) / 2 ~= m**2 / 2
        n_list = int(np.ceil(np.sqrt(2 * n_pool)))
        values = np.random.random(n_list)
    else:
        values = np.random.random(2 * n_pool).reshape((n_pool, 2))
    pool = Pool2D(values)
    test_values = pool.values[np.random.choice(range(len(pool.values)), n_test)] + \
        r_perturb * np.random.random(2 * n_test).reshape((n_test, 2))
    # At least a duplicate, for robustness
    test_values[-1] = test_values[0]
    indices = pool.find_indices(test_values)
    assert(np.all(np.abs(test_values - pool[indices]) < a_tol_test))


@flaky(max_runs=max_runs, min_passes=1)
def test_pool2d_from_list():
    test_pool2d(from_list=True)


def test_pool2d_fail():
    values = np.random.random(2)
    pool = Pool1D(values)
    test_values = [2, 2]  # out of range
    with pytest.raises(ValueError):
        pool.find_indices(test_values)



from cobaya.likelihoods.base_classes import DES


class joint(DES):
    r"""
    Combination of galaxy clustering and weak lensing data from the first year of the
    Dark Energy Survey (DES Y1) \cite{Abbott:2017wau}.
    """
    bibtex_file = 'des_y1.bibtex'



from cobaya.likelihoods.base_classes import BAO


class sdss_dr16_baoplus_lrg(BAO):
    r"""
    Likelihood of the power spectrum LRG BAO of SDSS DR16 \cite{Alam:2020sor}.
    As recommended in the official data release, it combines measurements
    of LRGs from eBOSS DR16 and BOSS DR12.
    """
    pass
